from __future__ import annotations

import json
import os
import uuid
from contextlib import AbstractContextManager
from contextvars import ContextVar
from enum import Enum
from types import TracebackType
from typing import List, Literal, Optional, Type, TypedDict

from opentelemetry.trace import Span, Status, StatusCode
from sqlalchemy.orm import Session

from dbos._utils import GlobalParams

from ._logger import dbos_logger
from ._request import Request
from ._tracer import dbos_tracer


# These are used to tag OTel traces
class OperationType(Enum):
    HANDLER = "handler"
    WORKFLOW = "workflow"
    TRANSACTION = "transaction"
    STEP = "step"
    PROCEDURE = "procedure"


OperationTypes = Literal["handler", "workflow", "transaction", "step", "procedure"]


# Keys must be the same as in TypeScript Transact
class TracedAttributes(TypedDict, total=False):
    name: str
    operationUUID: Optional[str]
    operationType: Optional[OperationTypes]
    requestID: Optional[str]
    requestIP: Optional[str]
    requestURL: Optional[str]
    requestMethod: Optional[str]
    applicationID: Optional[str]
    applicationVersion: Optional[str]
    executorID: Optional[str]
    authenticatedUser: Optional[str]
    authenticatedUserRoles: Optional[str]
    authenticatedUserAssumedRole: Optional[str]


class DBOSContext:
    def __init__(self) -> None:
        self.executor_id = GlobalParams.executor_id
        self.app_id = os.environ.get("DBOS__APPID", "")

        self.logger = dbos_logger

        self.request: Optional["Request"] = None

        self.id_assigned_for_next_workflow: str = ""
        self.is_within_set_workflow_id_block: bool = False

        self.parent_workflow_id: str = ""
        self.parent_workflow_fid: int = -1
        self.workflow_id: str = ""
        self.function_id: int = -1

        self.curr_step_function_id: int = -1
        self.curr_tx_function_id: int = -1
        self.sql_session: Optional[Session] = None
        self.spans: list[Span] = []

        self.authenticated_user: Optional[str] = None
        self.authenticated_roles: Optional[List[str]] = None
        self.assumed_role: Optional[str] = None

    def create_child(self) -> DBOSContext:
        rv = DBOSContext()
        rv.logger = self.logger
        rv.id_assigned_for_next_workflow = self.id_assigned_for_next_workflow
        self.id_assigned_for_next_workflow = ""
        rv.is_within_set_workflow_id_block = self.is_within_set_workflow_id_block
        rv.parent_workflow_id = self.workflow_id
        rv.parent_workflow_fid = self.function_id
        rv.authenticated_user = self.authenticated_user
        rv.authenticated_roles = (
            self.authenticated_roles[:]
            if self.authenticated_roles is not None
            else None
        )
        rv.request = self.request
        rv.assumed_role = self.assumed_role
        return rv

    def assign_workflow_id(self) -> str:
        if len(self.id_assigned_for_next_workflow) > 0:
            wfid = self.id_assigned_for_next_workflow
        else:
            if self.is_within_set_workflow_id_block:
                self.logger.warning(
                    f"Multiple workflows started in the same SetWorkflowID block. Only the first workflow is assigned the specified workflow ID; subsequent workflows will use a generated workflow ID."
                )
            wfid = str(uuid.uuid4())
        return wfid

    def start_workflow(
        self,
        wfid: Optional[str],
        attributes: TracedAttributes,
        is_temp_workflow: bool = False,
    ) -> None:
        if wfid is None or len(wfid) == 0:
            wfid = self.assign_workflow_id()
            self.id_assigned_for_next_workflow = ""
        self.workflow_id = wfid
        self.function_id = 0
        if not is_temp_workflow:
            self._start_span(attributes)

    def end_workflow(
        self, exc_value: Optional[BaseException], is_temp_workflow: bool = False
    ) -> None:
        self.workflow_id = ""
        self.function_id = -1
        if not is_temp_workflow:
            self._end_span(exc_value)

    def is_within_workflow(self) -> bool:
        return len(self.workflow_id) > 0

    def is_workflow(self) -> bool:
        return (
            len(self.workflow_id) > 0
            and not self.is_step()
            and not self.is_transaction()
        )

    def is_transaction(self) -> bool:
        return self.sql_session is not None

    def is_step(self) -> bool:
        return self.curr_step_function_id >= 0

    def start_step(
        self,
        fid: int,
        attributes: TracedAttributes,
    ) -> None:
        self.curr_step_function_id = fid
        self._start_span(attributes)

    def end_step(self, exc_value: Optional[BaseException]) -> None:
        self.curr_step_function_id = -1
        self._end_span(exc_value)

    def start_transaction(
        self, ses: Session, fid: int, attributes: TracedAttributes
    ) -> None:
        self.sql_session = ses
        self.curr_tx_function_id = fid
        self._start_span(attributes)

    def end_transaction(self, exc_value: Optional[BaseException]) -> None:
        self.sql_session = None
        self.curr_tx_function_id = -1
        self._end_span(exc_value)

    def start_handler(self, attributes: TracedAttributes) -> None:
        self._start_span(attributes)

    def end_handler(self, exc_value: Optional[BaseException]) -> None:
        self._end_span(exc_value)

    def get_current_span(self) -> Span:
        return self.spans[-1]

    def _start_span(self, attributes: TracedAttributes) -> None:
        attributes["operationUUID"] = (
            self.workflow_id if len(self.workflow_id) > 0 else None
        )
        attributes["authenticatedUser"] = self.authenticated_user
        attributes["authenticatedUserRoles"] = (
            json.dumps(self.authenticated_roles)
            if self.authenticated_roles is not None
            else ""
        )
        attributes["authenticatedUserAssumedRole"] = self.assumed_role
        span = dbos_tracer.start_span(
            attributes, parent=self.spans[-1] if len(self.spans) > 0 else None
        )
        self.spans.append(span)

    def _end_span(self, exc_value: Optional[BaseException]) -> None:
        if exc_value is None:
            self.spans[-1].set_status(Status(StatusCode.OK))
        else:
            self.spans[-1].set_status(
                Status(StatusCode.ERROR, description=str(exc_value))
            )
        dbos_tracer.end_span(self.spans.pop())

    def set_authentication(
        self, user: Optional[str], roles: Optional[List[str]]
    ) -> None:
        self.authenticated_user = user
        self.authenticated_roles = roles
        if user is not None and len(self.spans) > 0:
            self.spans[-1].set_attribute("authenticatedUser", user)
            self.spans[-1].set_attribute(
                "authenticatedUserRoles", json.dumps(roles) if roles is not None else ""
            )


##############################################################
##### Low-level context management (using contextvars)
##############################################################


_dbos_context_var: ContextVar[Optional[DBOSContext]] = ContextVar(
    "dbos_context", default=None
)


def _set_local_dbos_context(ctx: Optional[DBOSContext]) -> None:
    _dbos_context_var.set(ctx)


def _clear_local_dbos_context() -> None:
    _dbos_context_var.set(None)


def get_local_dbos_context() -> Optional[DBOSContext]:
    return _dbos_context_var.get()


def assert_current_dbos_context() -> DBOSContext:
    rv = get_local_dbos_context()
    assert rv, "No DBOS context found"
    return rv


##############################################################
##### High-level context management  (using contextlib)
##############################################################


class DBOSContextEnsure:
    def __init__(self) -> None:
        self.created_ctx = False

    def __enter__(self) -> DBOSContext:
        # Code to create a basic context
        ctx = get_local_dbos_context()
        if ctx is None:
            self.created_ctx = True
            _set_local_dbos_context(DBOSContext())
        return assert_current_dbos_context()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        # Code to clean up the basic context if we created it
        if self.created_ctx:
            _clear_local_dbos_context()
        return False  # Did not handle


class DBOSContextSwap:
    def __init__(self, ctx: DBOSContext) -> None:
        self.next_ctx = ctx
        self.prev_ctx: Optional[DBOSContext] = None

    def __enter__(self) -> DBOSContextSwap:
        self.prev_ctx = get_local_dbos_context()
        _set_local_dbos_context(self.next_ctx)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        assert get_local_dbos_context() == self.next_ctx
        _set_local_dbos_context(self.prev_ctx)
        return False  # Did not handle


class SetWorkflowID:
    """
    Set the workflow ID to be used for the enclosed workflow invocation. Note: Only the first workflow will be started with the specified workflow ID within a `with SetWorkflowID` block.

    Typical Usage
        ```
        with SetWorkflowID(<workflow ID>):
            result = workflow_function(...)
        ```

        or
        ```
        with SetWorkflowID(<workflow ID>):
            wf_handle = start_workflow(workflow_function, ...)
        ```
    """

    def __init__(self, wfid: str) -> None:
        self.created_ctx = False
        self.wfid = wfid

    def __enter__(self) -> SetWorkflowID:
        # Code to create a basic context
        ctx = get_local_dbos_context()
        if ctx is None:
            self.created_ctx = True
            _set_local_dbos_context(DBOSContext())
        ctx = assert_current_dbos_context()
        ctx.id_assigned_for_next_workflow = self.wfid
        ctx.is_within_set_workflow_id_block = True
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        # Code to clean up the basic context if we created it
        assert_current_dbos_context().is_within_set_workflow_id_block = False
        if self.created_ctx:
            _clear_local_dbos_context()
        return False  # Did not handle


class EnterDBOSWorkflow(AbstractContextManager[DBOSContext, Literal[False]]):
    def __init__(self, attributes: TracedAttributes) -> None:
        self.created_ctx = False
        self.attributes = attributes
        self.is_temp_workflow = attributes["name"] == "temp_wf"

    def __enter__(self) -> DBOSContext:
        # Code to create a basic context
        ctx = get_local_dbos_context()
        if ctx is None:
            self.created_ctx = True
            ctx = DBOSContext()
            _set_local_dbos_context(ctx)
        assert not ctx.is_within_workflow()
        ctx.start_workflow(
            None, self.attributes, self.is_temp_workflow
        )  # Will get from the context's next workflow ID
        return ctx

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        ctx = assert_current_dbos_context()
        assert ctx.is_within_workflow()
        ctx.end_workflow(exc_value, self.is_temp_workflow)
        # Code to clean up the basic context if we created it
        if self.created_ctx:
            _clear_local_dbos_context()
        return False  # Did not handle


class EnterDBOSChildWorkflow(AbstractContextManager[DBOSContext, Literal[False]]):
    def __init__(self, attributes: TracedAttributes) -> None:
        self.parent_ctx: Optional[DBOSContext] = None
        self.child_ctx: Optional[DBOSContext] = None
        self.attributes = attributes

    def __enter__(self) -> DBOSContext:
        ctx = assert_current_dbos_context()
        self.parent_ctx = ctx
        assert ctx.is_workflow()  # Is in a workflow and not in a step
        ctx.function_id += 1
        if len(ctx.id_assigned_for_next_workflow) == 0:
            ctx.id_assigned_for_next_workflow = (
                ctx.workflow_id + "-" + str(ctx.function_id)
            )
        self.child_ctx = ctx.create_child()
        _set_local_dbos_context(self.child_ctx)
        self.child_ctx.start_workflow(None, attributes=self.attributes)
        return self.child_ctx

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        ctx = assert_current_dbos_context()
        assert ctx.is_within_workflow()
        ctx.end_workflow(exc_value)
        # Return to parent ctx
        assert self.parent_ctx
        _set_local_dbos_context(self.parent_ctx)
        return False  # Did not handle


class EnterDBOSStep:
    def __init__(
        self,
        attributes: TracedAttributes,
    ) -> None:
        self.attributes = attributes

    def __enter__(self) -> DBOSContext:
        ctx = assert_current_dbos_context()
        assert ctx.is_workflow()
        ctx.function_id += 1
        ctx.start_step(ctx.function_id, attributes=self.attributes)
        return ctx

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        ctx = assert_current_dbos_context()
        assert ctx.is_step()
        ctx.end_step(exc_value)
        return False  # Did not handle


class EnterDBOSTransaction:
    def __init__(self, sqls: Session, attributes: TracedAttributes) -> None:
        self.sqls = sqls
        self.attributes = attributes

    def __enter__(self) -> DBOSContext:
        ctx = assert_current_dbos_context()
        assert ctx.is_workflow()
        ctx.function_id += 1
        ctx.start_transaction(self.sqls, ctx.function_id, attributes=self.attributes)
        return ctx

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        ctx = assert_current_dbos_context()
        assert ctx.is_transaction()
        ctx.end_transaction(exc_value)
        return False  # Did not handle


class EnterDBOSHandler:
    def __init__(self, attributes: TracedAttributes) -> None:
        self.created_ctx = False
        self.attributes = attributes

    def __enter__(self) -> EnterDBOSHandler:
        # Code to create a basic context
        ctx = get_local_dbos_context()
        if ctx is None:
            self.created_ctx = True
            _set_local_dbos_context(DBOSContext())
        ctx = assert_current_dbos_context()
        ctx.start_handler(self.attributes)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        ctx = assert_current_dbos_context()
        ctx.end_handler(exc_value)
        # Code to clean up the basic context if we created it
        if self.created_ctx:
            _clear_local_dbos_context()
        return False  # Did not handle


class DBOSContextSetAuth(DBOSContextEnsure):
    def __init__(self, user: Optional[str], roles: Optional[List[str]]) -> None:
        self.created_ctx = False
        self.user = user
        self.roles = roles
        self.prev_user: Optional[str] = None
        self.prev_roles: Optional[List[str]] = None

    def __enter__(self) -> DBOSContext:
        ctx = get_local_dbos_context()
        if ctx is None:
            self.created_ctx = True
            _set_local_dbos_context(DBOSContext())
        ctx = assert_current_dbos_context()
        self.prev_user = ctx.authenticated_user
        self.prev_roles = ctx.authenticated_roles
        ctx.set_authentication(self.user, self.roles)
        return ctx

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        ctx = assert_current_dbos_context()
        ctx.set_authentication(self.prev_user, self.prev_roles)
        # Clean up the basic context if we created it
        if self.created_ctx:
            _clear_local_dbos_context()
        return False  # Did not handle


class DBOSAssumeRole:
    def __init__(self, assume_role: Optional[str]) -> None:
        self.prior_role: Optional[str] = None
        self.assume_role = assume_role

    def __enter__(self) -> DBOSAssumeRole:
        ctx = assert_current_dbos_context()
        self.prior_role = ctx.assumed_role
        if self.assume_role is not None:
            ctx.assumed_role = self.assume_role
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        ctx = assert_current_dbos_context()
        if self.assume_role is not None:
            assert ctx.assumed_role == self.assume_role
        ctx.assumed_role = self.prior_role
        return False  # Did not handle



import random
import threading
import time
from typing import NoReturn

import pytest
from confluent_kafka import KafkaError, Producer

from dbos import DBOS, KafkaMessage

# These tests require local Kafka to run.
# Without it, they're automatically skipped.
# Here's a docker-compose script you can use to set up local Kafka:

# version: "3.7"
# services:
#   broker:
#       image: bitnami/kafka:latest
#       hostname: broker
#       container_name: broker
#       ports:
#         - '9092:9092'
#       environment:
#         KAFKA_CFG_NODE_ID: 1
#         KAFKA_CFG_LISTENER_SECURITY_PROTOCOL_MAP: 'CONTROLLER:PLAINTEXT,PLAINTEXT:PLAINTEXT,PLAINTEXT_HOST:PLAINTEXT'
#         KAFKA_CFG_ADVERTISED_LISTENERS: 'PLAINTEXT_HOST://localhost:9092,PLAINTEXT://broker:19092'
#         KAFKA_CFG_PROCESS_ROLES: 'broker,controller'
#         KAFKA_CFG_CONTROLLER_QUORUM_VOTERS: '1@broker:29093'
#         KAFKA_CFG_LISTENERS: 'CONTROLLER://:29093,PLAINTEXT_HOST://:9092,PLAINTEXT://:19092'
#         KAFKA_CFG_INTER_BROKER_LISTENER_NAME: 'PLAINTEXT'
#         KAFKA_CFG_CONTROLLER_LISTENER_NAMES: 'CONTROLLER'


NUM_EVENTS = 3


def send_test_messages(server: str, topic: str) -> bool:

    try:

        def on_error(err: KafkaError) -> NoReturn:
            raise Exception(err)

        producer = Producer({"bootstrap.servers": server, "error_cb": on_error})

        for i in range(NUM_EVENTS):
            producer.produce(
                topic, key=f"test message key {i}", value=f"test message value {i}"
            )

        producer.poll(10)
        producer.flush(10)
        return True
    except Exception as e:
        return False
    finally:
        pass


def test_kafka(dbos: DBOS) -> None:
    event = threading.Event()
    kafka_count = 0
    server = "localhost:9092"
    topic = f"dbos-kafka-{random.randrange(1_000_000_000)}"

    if not send_test_messages(server, topic):
        pytest.skip("Kafka not available")

    @DBOS.kafka_consumer(
        {
            "bootstrap.servers": server,
            "group.id": "dbos-test",
            "auto.offset.reset": "earliest",
        },
        [topic],
    )
    @DBOS.workflow()
    def test_kafka_workflow(msg: KafkaMessage) -> None:
        nonlocal kafka_count
        kafka_count += 1
        assert b"test message key" in msg.key  # type: ignore
        assert b"test message value" in msg.value  # type: ignore
        print(msg)
        if kafka_count == 3:
            event.set()

    wait = event.wait(timeout=10)
    assert wait
    assert kafka_count == 3


def test_kafka_in_order(dbos: DBOS) -> None:
    event = threading.Event()
    kafka_count = 0
    server = "localhost:9092"
    topic = f"dbos-kafka-{random.randrange(1_000_000_000)}"

    if not send_test_messages(server, topic):
        pytest.skip("Kafka not available")

    @DBOS.kafka_consumer(
        {
            "bootstrap.servers": server,
            "group.id": "dbos-test",
            "auto.offset.reset": "earliest",
        },
        [topic],
        in_order=True,
    )
    @DBOS.workflow()
    def test_kafka_workflow(msg: KafkaMessage) -> None:
        time.sleep(random.uniform(0, 2))
        nonlocal kafka_count
        kafka_count += 1
        assert f"test message key {kafka_count - 1}".encode() == msg.key
        print(msg)
        if kafka_count == 3:
            event.set()

    wait = event.wait(timeout=15)
    assert wait
    assert kafka_count == 3
    time.sleep(2)  # Wait for things to clean up


def test_kafka_no_groupid(dbos: DBOS) -> None:
    event = threading.Event()
    kafka_count = 0
    server = "localhost:9092"
    topic1 = f"dbos-kafka-{random.randrange(1_000_000_000, 2_000_000_000)}"
    topic2 = f"dbos-kafka-{random.randrange(2_000_000_000, 3_000_000_000)}"

    if not send_test_messages(server, topic1):
        pytest.skip("Kafka not available")

    if not send_test_messages(server, topic2):
        pytest.skip("Kafka not available")

    @DBOS.kafka_consumer(
        {
            "bootstrap.servers": server,
            "auto.offset.reset": "earliest",
        },
        [topic1, topic2],
    )
    @DBOS.workflow()
    def test_kafka_workflow(msg: KafkaMessage) -> None:
        nonlocal kafka_count
        kafka_count += 1
        assert b"test message key" in msg.key  # type: ignore
        assert b"test message value" in msg.value  # type: ignore
        print(msg)
        if kafka_count == 6:
            event.set()

    wait = event.wait(timeout=10)
    assert wait
    assert kafka_count == 6



import os
import subprocess
import sys
import time
from os import path

import pytest

# Public API
from dbos import DBOS, SetWorkflowID, WorkflowHandle

# Private API used because this is a test
from dbos._context import DBOSContextEnsure, assert_current_dbos_context
from tests.conftest import default_config


def test_dbos_singleton(cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    # Simulate an app that does some imports of its own code, then defines DBOS,
    #    then imports more
    from tests.classdefs import DBOSSendRecv, DBOSTestClass, DBOSTestRoles

    dbos: DBOS = DBOS(config=default_config())

    from tests.more_classdefs import DBOSWFEvents, wfFunc

    DBOS.launch()  # Usually framework (fastapi) does this via lifecycle event

    # Basics
    with SetWorkflowID("wfid"):
        res = wfFunc("f")
    assert res == "f1"

    res = DBOSTestClass.test_workflow_cls("a", "b")
    assert res == "b1a"

    inst = DBOSTestClass()
    res = inst.test_workflow("c", "d")
    assert res == "d1c"

    wh = DBOS.start_workflow(DBOSTestClass.test_workflow_cls, "a", "b")
    assert wh.get_result() == "b1a"

    wh = DBOS.start_workflow(inst.test_workflow, "c", "d")
    assert wh.get_result() == "d1c"
    stati = DBOS.get_workflow_status(wh.get_workflow_id())
    assert stati
    assert stati.config_name == "myconfig"
    assert stati.class_name == "DBOSTestClass"
    wfhr: WorkflowHandle[str] = DBOS.retrieve_workflow(wh.get_workflow_id())
    assert wfhr.workflow_id == wh.get_workflow_id()

    # Roles

    with DBOSContextEnsure():
        with pytest.raises(Exception) as exc_info:
            DBOSTestRoles.greetfunc("Nobody")
        assert "DBOS Error 8" in str(exc_info.value)

        ctx = assert_current_dbos_context()
        ctx.authenticated_roles = ["user", "admin"]
        res = inst.test_func_admin("admin")
        assert res == "myconfig:admin"

        assert DBOSTestRoles.greetfunc("user") == "Hello user"

    # Send / Recv
    dest_uuid = str("sruuid1")

    with SetWorkflowID(dest_uuid):
        handle = dbos.start_workflow(DBOSSendRecv.test_recv_workflow, "testtopic")
        assert handle.get_workflow_id() == dest_uuid

    send_uuid = str("sruuid2")
    with SetWorkflowID(send_uuid):
        res = DBOSSendRecv.test_send_workflow(handle.get_workflow_id(), "testtopic")
        assert res == dest_uuid

    begin_time = time.time()
    assert handle.get_result() == "test2-test1-test3"
    duration = time.time() - begin_time
    assert duration < 3.0

    # Events
    wfuuid = str("sendwf1")
    with SetWorkflowID(wfuuid):
        DBOSWFEvents.test_setevent_workflow()
    with SetWorkflowID(wfuuid):
        DBOSWFEvents.test_setevent_workflow()

    value1 = DBOSWFEvents.test_getevent_workflow(wfuuid, "key1")
    assert value1 == "value1"

    value2 = DBOSWFEvents.test_getevent_workflow(wfuuid, "key2")
    assert value2 == "value2"

    # Run getEvent outside of a workflow
    value1 = DBOS.get_event(wfuuid, "key1")
    assert value1 == "value1"

    value2 = DBOS.get_event(wfuuid, "key2")
    assert value2 == "value2"

    begin_time = time.time()
    value3 = DBOSWFEvents.test_getevent_workflow(wfuuid, "key3")
    assert value3 is None
    duration = time.time() - begin_time
    assert duration < 1  # None is from the event not from the timeout

    DBOS.destroy()


def test_dbos_singleton_negative(cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    # Simulate an app that does some imports of its own code, then defines DBOS,
    #    then imports more
    from tests.classdefs import DBOSTestClass

    dbos: DBOS = DBOS(config=default_config())

    # Don't initialize DBOS twice
    with pytest.raises(Exception) as exc_info:
        DBOS(config=default_config())
    assert "conflicting configuration" in str(exc_info.value)

    # Something should have launched
    with pytest.raises(Exception) as exc_info:
        DBOSTestClass.test_workflow_cls("a", "b")
    assert "launch" in str(exc_info.value)

    DBOS.destroy()


config_string = """name: test-app
language: python
database:
  hostname: localhost
  port: 5432
  username: postgres
  password: ${PGPASSWORD}
  app_db_name: dbostestpy
runtimeConfig:
  start:
    - python3 main.py
application:
  service_url: 'https://service.org'
  service_config:
    port: 80
    user: "user"
    password: "password"
"""


def test_config_before_singleton(cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    # Write the config to a text file for the moment
    with open("dbos-config.yaml", "w") as file:
        file.write(config_string)

    try:
        # Simulate an app that does some imports of its own code, then defines DBOS,
        #    then imports more
        from tests.classdefs import DBOSTestClass

        x = DBOS.config.get("language")
        assert x == "python"
        y = DBOS.config["language"]
        assert y == "python"
        url = DBOS.config["application"]["service_url"]
        assert url == "https://service.org"
        port = DBOS.config["application"]["service_config"]["port"]
        assert port == 80

        # This is OK, it meant load_config anyway
        dbos: DBOS = DBOS()

    finally:
        # Initialize singleton
        DBOS.destroy()  # In case of other tests leaving it
        os.remove("dbos-config.yaml")


def test_config_before_singleton_negative(cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    # Write the config to a text file for the moment
    with open("dbos-config.yaml", "w") as file:
        file.write(config_string)

    try:
        x = DBOS.config.get("language")
        assert x == "python"

        # Not OK, config already loaded in the default way
        with pytest.raises(Exception) as exc_info:
            DBOS(config=default_config())
        assert "configured multiple" in str(exc_info.value)
    finally:
        DBOS.destroy()
        os.remove("dbos-config.yaml")


def test_dbos_atexit_no_dbos(cleanup_test_databases: None) -> None:
    # Run the .py as a separate process
    result = subprocess.run(
        [sys.executable, path.join("tests", "atexit_no_ctor.py")],
        capture_output=True,
        text=True,
    )

    # Assert that the output contains the warning message
    assert "DBOS exiting; functions were registered" in result.stdout


def test_dbos_atexit_no_launch(cleanup_test_databases: None) -> None:
    # Run the .py as a separate process
    result = subprocess.run(
        [sys.executable, path.join("tests", "atexit_no_launch.py")],
        capture_output=True,
        text=True,
    )

    # Assert that the output contains the warning message
    assert "DBOS exists but launch() was not called" in result.stdout



from pdm.backend.hooks.version import SCMVersion


def format_version(git_version: SCMVersion) -> str:
    """
    Format version into string.

    1. Release versions may only be published from release branches. Their version is a git tag.
    2. Preview versions are published from main. They are PEP440 alpha releases whose version is the
    next release version number followed by "a" followed by the number of commits since the last release.
    If the last release was 1.2.3 and there have been ten commits since, the preview version is 1.2.3a10
    3. Test versions are published from feature branches. They are PEP440 local versions tagged with a git hash.
    """
    assert git_version.branch is not None
    is_release = "release" in git_version.branch
    is_preview = git_version.branch == "main"

    next_version = guess_next_version(str(git_version.version))

    if git_version.distance is None:
        if is_release:
            version = str(git_version.version)
        elif is_preview:
            version = f"{next_version}a0"
        else:
            version = f"{next_version}a0+{git_version.node}"
    else:
        if is_release:
            raise Exception(
                f"Release branches may only publish tagged releases. Distance: {git_version.distance}"
            )
        elif is_preview:
            version = f"{next_version}a{git_version.distance}"
        else:
            version = f"{next_version}a{git_version.distance}+{git_version.node}"

    return version


def guess_next_version(version_number: str) -> str:
    major, minor, patch = map(int, version_number.split("."))
    minor += 1
    return f"{major}.{minor}.{patch}"



from typing import List, Optional, cast

from . import _serialization
from ._sys_db import (
    GetQueuedWorkflowsInput,
    GetWorkflowsInput,
    GetWorkflowsOutput,
    SystemDatabase,
    WorkflowStatuses,
)


class WorkflowInformation:
    workflow_id: str
    status: WorkflowStatuses
    workflow_name: str
    workflow_class_name: Optional[str]
    workflow_config_name: Optional[str]
    authenticated_user: Optional[str]
    assumed_role: Optional[str]
    authenticated_roles: Optional[str]  # JSON list of roles.
    input: Optional[_serialization.WorkflowInputs]  # JSON (jsonpickle)
    output: Optional[str] = None  # JSON (jsonpickle)
    request: Optional[str]  # JSON (jsonpickle)
    error: Optional[str] = None  # JSON (jsonpickle)
    created_at: Optional[int]  # Unix epoch timestamp in ms
    updated_at: Optional[int]  # Unix epoch timestamp in ms
    queue_name: Optional[str]
    executor_id: Optional[str]
    app_version: Optional[str]
    app_id: Optional[str]
    recovery_attempts: Optional[int]


def list_workflows(
    sys_db: SystemDatabase,
    *,
    workflow_ids: Optional[List[str]] = None,
    user: Optional[str] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    status: Optional[str] = None,
    request: bool = False,
    app_version: Optional[str] = None,
    name: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    sort_desc: bool = False,
) -> List[WorkflowInformation]:
    input = GetWorkflowsInput()
    input.workflow_ids = workflow_ids
    input.authenticated_user = user
    input.start_time = start_time
    input.end_time = end_time
    if status is not None:
        input.status = cast(WorkflowStatuses, status)
    input.application_version = app_version
    input.limit = limit
    input.name = name
    input.offset = offset
    input.sort_desc = sort_desc

    output: GetWorkflowsOutput = sys_db.get_workflows(input)
    infos: List[WorkflowInformation] = []
    for workflow_id in output.workflow_uuids:
        info = get_workflow(sys_db, workflow_id, request)  # Call the method for each ID
        if info is not None:
            infos.append(info)
    return infos


def list_queued_workflows(
    sys_db: SystemDatabase,
    *,
    limit: Optional[int] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    queue_name: Optional[str] = None,
    status: Optional[str] = None,
    name: Optional[str] = None,
    request: bool = False,
    offset: Optional[int] = None,
    sort_desc: bool = False,
) -> List[WorkflowInformation]:
    input: GetQueuedWorkflowsInput = {
        "queue_name": queue_name,
        "start_time": start_time,
        "end_time": end_time,
        "status": status,
        "limit": limit,
        "name": name,
        "offset": offset,
        "sort_desc": sort_desc,
    }
    output: GetWorkflowsOutput = sys_db.get_queued_workflows(input)
    infos: List[WorkflowInformation] = []
    for workflow_id in output.workflow_uuids:
        info = get_workflow(sys_db, workflow_id, request)  # Call the method for each ID
        if info is not None:
            infos.append(info)
    return infos


def get_workflow(
    sys_db: SystemDatabase, workflowUUID: str, getRequest: bool
) -> Optional[WorkflowInformation]:

    info = sys_db.get_workflow_status(workflowUUID)
    if info is None:
        return None

    winfo = WorkflowInformation()

    winfo.workflow_id = workflowUUID
    winfo.status = info["status"]
    winfo.workflow_name = info["name"]
    winfo.workflow_class_name = info["class_name"]
    winfo.workflow_config_name = info["config_name"]
    winfo.authenticated_user = info["authenticated_user"]
    winfo.assumed_role = info["assumed_role"]
    winfo.authenticated_roles = info["authenticated_roles"]
    winfo.request = info["request"]
    winfo.created_at = info["created_at"]
    winfo.updated_at = info["updated_at"]
    winfo.queue_name = info["queue_name"]
    winfo.executor_id = info["executor_id"]
    winfo.app_version = info["app_version"]
    winfo.app_id = info["app_id"]
    winfo.recovery_attempts = info["recovery_attempts"]

    # no input field
    input_data = sys_db.get_workflow_inputs(workflowUUID)
    if input_data is not None:
        winfo.input = input_data

    if info.get("status") == "SUCCESS":
        result = sys_db.await_workflow_result(workflowUUID)
        winfo.output = result
    elif info.get("status") == "ERROR":
        try:
            sys_db.await_workflow_result(workflowUUID)
        except Exception as e:
            winfo.error = str(e)

    if not getRequest:
        winfo.request = None

    return winfo



from sqlalchemy import (
    BigInteger,
    Column,
    Index,
    Integer,
    MetaData,
    PrimaryKeyConstraint,
    Table,
    Text,
    text,
)


class ApplicationSchema:
    schema = "dbos"
    metadata_obj = MetaData(schema=schema)

    transaction_outputs = Table(
        "transaction_outputs",
        metadata_obj,
        Column("workflow_uuid", Text),
        Column("function_id", Integer),
        Column("output", Text, nullable=True),
        Column("error", Text, nullable=True),
        Column("txn_id", Text, nullable=True),
        Column("txn_snapshot", Text),
        Column("executor_id", Text, nullable=True),
        Column(
            "created_at",
            BigInteger,
            nullable=False,
            server_default=text("(EXTRACT(epoch FROM now()) * 1000::numeric)::bigint"),
        ),
        Index("transaction_outputs_created_at_index", "created_at"),
        PrimaryKeyConstraint("workflow_uuid", "function_id"),
    )



import types
from typing import Any, Dict, Tuple, TypedDict

import jsonpickle  # type: ignore


class WorkflowInputs(TypedDict):
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]


def _validate_item(data: Any) -> None:
    if isinstance(data, (types.FunctionType, types.MethodType)):
        raise TypeError("Serialized data item should not be a function")


def serialize(data: Any) -> str:
    """Serialize an object to a JSON string using jsonpickle."""
    _validate_item(data)
    encoded_data: str = jsonpickle.encode(data, unpicklable=True)
    return encoded_data


def serialize_args(data: WorkflowInputs) -> str:
    """Serialize args to a JSON string using jsonpickle."""
    arg: Any
    for arg in data["args"]:
        _validate_item(arg)
    for arg in data["kwargs"].values():
        _validate_item(arg)
    encoded_data: str = jsonpickle.encode(data, unpicklable=True)
    return encoded_data


def serialize_exception(data: Exception) -> str:
    """Serialize an Exception object to a JSON string using jsonpickle."""
    encoded_data: str = jsonpickle.encode(data, unpicklable=True)
    return encoded_data


def deserialize(serialized_data: str) -> Any:
    """Deserialize a JSON string back to a Python object using jsonpickle."""
    return jsonpickle.decode(serialized_data)


def deserialize_args(serialized_data: str) -> WorkflowInputs:
    """Deserialize a JSON string back to a Python object list using jsonpickle."""
    args: WorkflowInputs = jsonpickle.decode(serialized_data)
    return args


def deserialize_exception(serialized_data: str) -> Exception:
    """Deserialize JSON string back to a Python Exception using jsonpickle."""
    upo: Exception = jsonpickle.decode(serialized_data)
    return upo



# type: ignore
# This code was based on code from croniter: https://pypi.org/project/croniter/
"""
Copyright (C) 2010-2012 Matsumoto Taichi.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import absolute_import, division, print_function

import binascii
import calendar
import copy
import datetime
import inspect
import math
import random
import re
import sys
import traceback as _traceback
from collections import OrderedDict
from time import time

from dateutil.relativedelta import relativedelta
from dateutil.tz import tzutc

M_ALPHAS = {
    "jan": 1,
    "feb": 2,
    "mar": 3,
    "apr": 4,
    "may": 5,
    "jun": 6,
    "jul": 7,
    "aug": 8,
    "sep": 9,
    "oct": 10,
    "nov": 11,
    "dec": 12,
}
DOW_ALPHAS = {"sun": 0, "mon": 1, "tue": 2, "wed": 3, "thu": 4, "fri": 5, "sat": 6}
ALPHAS = {}
for i in M_ALPHAS, DOW_ALPHAS:
    ALPHAS.update(i)
del i
step_search_re = re.compile(r"^([^-]+)-([^-/]+)(/(\d+))?$")
only_int_re = re.compile(r"^\d+$")

WEEKDAYS = "|".join(DOW_ALPHAS.keys())
MONTHS = "|".join(M_ALPHAS.keys())
star_or_int_re = re.compile(r"^(\d+|\*)$")
special_dow_re = re.compile(
    (r"^(?P<pre>((?P<he>(({WEEKDAYS})(-({WEEKDAYS}))?)").format(WEEKDAYS=WEEKDAYS)
    + (r"|(({MONTHS})(-({MONTHS}))?)|\w+)#)|l)(?P<last>\d+)$").format(MONTHS=MONTHS)
)
re_star = re.compile("[*]")
hash_expression_re = re.compile(
    r"^(?P<hash_type>h|r)(\((?P<range_begin>\d+)-(?P<range_end>\d+)\))?(\/(?P<divisor>\d+))?$"
)
MINUTE_FIELD = 0
HOUR_FIELD = 1
DAY_FIELD = 2
MONTH_FIELD = 3
DOW_FIELD = 4
SECOND_FIELD = 5
YEAR_FIELD = 6
UNIX_FIELDS = (MINUTE_FIELD, HOUR_FIELD, DAY_FIELD, MONTH_FIELD, DOW_FIELD)
SECOND_FIELDS = (
    MINUTE_FIELD,
    HOUR_FIELD,
    DAY_FIELD,
    MONTH_FIELD,
    DOW_FIELD,
    SECOND_FIELD,
)
YEAR_FIELDS = (
    MINUTE_FIELD,
    HOUR_FIELD,
    DAY_FIELD,
    MONTH_FIELD,
    DOW_FIELD,
    SECOND_FIELD,
    YEAR_FIELD,
)
CRON_FIELDS = {
    "unix": UNIX_FIELDS,
    "second": SECOND_FIELDS,
    "year": YEAR_FIELDS,
    len(UNIX_FIELDS): UNIX_FIELDS,
    len(SECOND_FIELDS): SECOND_FIELDS,
    len(YEAR_FIELDS): YEAR_FIELDS,
}
UNIX_CRON_LEN = len(UNIX_FIELDS)
SECOND_CRON_LEN = len(SECOND_FIELDS)
YEAR_CRON_LEN = len(YEAR_FIELDS)
# retrocompat
VALID_LEN_EXPRESSION = set([a for a in CRON_FIELDS if isinstance(a, int)])
EXPRESSIONS = {}


def timedelta_to_seconds(td):
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6


def datetime_to_timestamp(d):
    if d.tzinfo is not None:
        d = d.replace(tzinfo=None) - d.utcoffset()

    return timedelta_to_seconds(d - datetime.datetime(1970, 1, 1))


def _get_caller_globals_and_locals():
    """
    Return the globals and locals of the calling frame.

    Is there an alternative to frame hacking here?
    """
    caller_frame = inspect.stack()[2]
    myglobals = caller_frame[0].f_globals
    mylocals = caller_frame[0].f_locals
    return myglobals, mylocals


class CroniterError(ValueError):
    """General top-level Croniter base exception."""

    pass


class CroniterBadTypeRangeError(TypeError):
    """Bad Type Range."""


class CroniterBadCronError(CroniterError):
    """Syntax, unknown value, or range error within a cron expression."""

    pass


class CroniterUnsupportedSyntaxError(CroniterBadCronError):
    """Valid cron syntax, but likely to produce inaccurate results."""

    # Extending CroniterBadCronError, which may be contridatory, but this allows
    # catching both errors with a single exception.  From a user perspective
    # these will likely be handled the same way.
    pass


class CroniterBadDateError(CroniterError):
    """Unable to find next/prev timestamp match."""

    pass


class CroniterNotAlphaError(CroniterBadCronError):
    """Cron syntax contains an invalid day or month abbreviation."""

    pass


class croniter(object):
    MONTHS_IN_YEAR = 12
    RANGES = ((0, 59), (0, 23), (1, 31), (1, 12), (0, 7), (0, 59), (1970, 2099))
    DAYS = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)

    ALPHACONV = (
        {},  # 0: min
        {},  # 1: hour
        {"l": "l"},  # 2: dom
        # 3: mon
        copy.deepcopy(M_ALPHAS),
        # 4: dow
        copy.deepcopy(DOW_ALPHAS),
        # 5: second
        {},
        # 6: year
        {},
    )

    LOWMAP = ({}, {}, {0: 1}, {0: 1}, {7: 0}, {}, {})

    LEN_MEANS_ALL = (60, 24, 31, 12, 7, 60, 130)

    bad_length = (
        "Exactly 5, 6 or 7 columns has to be specified for iterator " "expression."
    )

    def __init__(
        self,
        expr_format,
        start_time=None,
        ret_type=float,
        day_or=True,
        max_years_between_matches=None,
        is_prev=False,
        hash_id=None,
        implement_cron_bug=False,
        second_at_beginning=None,
        expand_from_start_time=False,
    ):
        self._ret_type = ret_type
        self._day_or = day_or
        self._implement_cron_bug = implement_cron_bug
        self.second_at_beginning = bool(second_at_beginning)
        self._expand_from_start_time = expand_from_start_time

        if hash_id:
            if not isinstance(hash_id, (bytes, str)):
                raise TypeError("hash_id must be bytes or UTF-8 string")
            if not isinstance(hash_id, bytes):
                hash_id = hash_id.encode("UTF-8")

        self._max_years_btw_matches_explicitly_set = (
            max_years_between_matches is not None
        )
        if not self._max_years_btw_matches_explicitly_set:
            max_years_between_matches = 50
        self._max_years_between_matches = max(int(max_years_between_matches), 1)

        if start_time is None:
            start_time = time()

        self.tzinfo = None

        self.start_time = None
        self.dst_start_time = None
        self.cur = None
        self.set_current(start_time, force=False)

        self.expanded, self.nth_weekday_of_month = self.expand(
            expr_format,
            hash_id=hash_id,
            from_timestamp=(
                self.dst_start_time if self._expand_from_start_time else None
            ),
            second_at_beginning=second_at_beginning,
        )
        self.fields = CRON_FIELDS[len(self.expanded)]
        self.expressions = EXPRESSIONS[(expr_format, hash_id, second_at_beginning)]
        self._is_prev = is_prev

    @classmethod
    def _alphaconv(cls, index, key, expressions):
        try:
            return cls.ALPHACONV[index][key]
        except KeyError:
            raise CroniterNotAlphaError(
                "[{0}] is not acceptable".format(" ".join(expressions))
            )

    def get_next(self, ret_type=None, start_time=None, update_current=True):
        if start_time and self._expand_from_start_time:
            raise ValueError(
                "start_time is not supported when using expand_from_start_time = True."
            )
        return self._get_next(
            ret_type or self._ret_type,
            start_time=start_time,
            is_prev=False,
            update_current=update_current,
        )

    def get_prev(self, ret_type=None, start_time=None, update_current=True):
        return self._get_next(
            ret_type or self._ret_type,
            start_time=start_time,
            is_prev=True,
            update_current=update_current,
        )

    def get_current(self, ret_type=None):
        ret_type = ret_type or self._ret_type
        if issubclass(ret_type, datetime.datetime):
            return self._timestamp_to_datetime(self.cur)
        return self.cur

    def set_current(self, start_time, force=True):
        if (force or (self.cur is None)) and start_time is not None:
            if isinstance(start_time, datetime.datetime):
                self.tzinfo = start_time.tzinfo
                start_time = self._datetime_to_timestamp(start_time)

            self.start_time = start_time
            self.dst_start_time = start_time
            self.cur = start_time
        return self.cur

    @classmethod
    def _datetime_to_timestamp(cls, d):
        """Convert a `datetime` object `d` into a UNIX timestamp."""
        return datetime_to_timestamp(d)

    def _timestamp_to_datetime(self, timestamp):
        """Convert a UNIX timestamp `timestamp` into a `datetime` object."""
        result = datetime.datetime.fromtimestamp(timestamp, tz=tzutc()).replace(
            tzinfo=None
        )
        if self.tzinfo:
            result = result.replace(tzinfo=tzutc()).astimezone(self.tzinfo)

        return result

    @classmethod
    def _timedelta_to_seconds(cls, td):
        """
        Convert a 'datetime.timedelta' object `td` into seconds contained in the duration.

        Note: We cannot use `timedelta.total_seconds()` because this is not
        supported by Python 2.6.
        """
        return timedelta_to_seconds(td)

    def _get_next(
        self, ret_type=None, start_time=None, is_prev=None, update_current=None
    ):
        if update_current is None:
            update_current = True
        self.set_current(start_time, force=True)
        if is_prev is None:
            is_prev = self._is_prev
        self._is_prev = is_prev
        expanded = self.expanded[:]
        nth_weekday_of_month = self.nth_weekday_of_month.copy()

        ret_type = ret_type or self._ret_type

        if not issubclass(ret_type, (float, datetime.datetime)):
            raise TypeError(
                "Invalid ret_type, only 'float' or 'datetime' " "is acceptable."
            )

        # exception to support day of month and day of week as defined in cron
        dom_dow_exception_processed = False
        if (
            expanded[DAY_FIELD][0] != "*" and expanded[DOW_FIELD][0] != "*"
        ) and self._day_or:
            # If requested, handle a bug in vixie cron/ISC cron where day_of_month and day_of_week form
            # an intersection (AND) instead of a union (OR) if either field is an asterisk or starts with an asterisk
            # (https://crontab.guru/cron-bug.html)
            if self._implement_cron_bug and (
                re_star.match(self.expressions[DAY_FIELD])
                or re_star.match(self.expressions[DOW_FIELD])
            ):
                # To produce a schedule identical to the cron bug, we'll bypass the code that
                # makes a union of DOM and DOW, and instead skip to the code that does an intersect instead
                pass
            else:
                bak = expanded[DOW_FIELD]
                expanded[DOW_FIELD] = ["*"]
                t1 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
                expanded[DOW_FIELD] = bak
                expanded[DAY_FIELD] = ["*"]

                t2 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
                if not is_prev:
                    result = t1 if t1 < t2 else t2
                else:
                    result = t1 if t1 > t2 else t2
                dom_dow_exception_processed = True

        if not dom_dow_exception_processed:
            result = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)

        # DST Handling for cron job spanning across days
        dtstarttime = self._timestamp_to_datetime(self.dst_start_time)
        dtstarttime_utcoffset = dtstarttime.utcoffset() or datetime.timedelta(0)
        dtresult = self._timestamp_to_datetime(result)
        lag = lag_hours = 0
        # do we trigger DST on next crontab (handle backward changes)
        dtresult_utcoffset = dtstarttime_utcoffset
        if dtresult and self.tzinfo:
            dtresult_utcoffset = dtresult.utcoffset()
            lag_hours = self._timedelta_to_seconds(dtresult - dtstarttime) / (60 * 60)
            lag = self._timedelta_to_seconds(dtresult_utcoffset - dtstarttime_utcoffset)
        hours_before_midnight = 24 - dtstarttime.hour
        if dtresult_utcoffset != dtstarttime_utcoffset:
            if (lag > 0 and abs(lag_hours) >= hours_before_midnight) or (
                lag < 0
                and ((3600 * abs(lag_hours) + abs(lag)) >= hours_before_midnight * 3600)
            ):
                dtresult_adjusted = dtresult - datetime.timedelta(seconds=lag)
                result_adjusted = self._datetime_to_timestamp(dtresult_adjusted)
                # Do the actual adjust only if the result time actually exists
                if (
                    self._timestamp_to_datetime(result_adjusted).tzinfo
                    == dtresult_adjusted.tzinfo
                ):
                    dtresult = dtresult_adjusted
                    result = result_adjusted
                self.dst_start_time = result
        if update_current:
            self.cur = result
        if issubclass(ret_type, datetime.datetime):
            result = dtresult
        return result

    # iterator protocol, to enable direct use of croniter
    # objects in a loop, like "for dt in croniter('5 0 * * *'): ..."
    # or for combining multiple croniters into single
    # dates feed using 'itertools' module
    def all_next(self, ret_type=None, start_time=None, update_current=None):
        """
        Generate all consecutive dates.

        Can be used instead of
        implicit call to __iter__, whenever non-default
        'ret_type' has to be specified.
        """
        # In a Python 3.7+ world:  contextlib.suppress and contextlib.nullcontext could be used instead
        try:
            while True:
                self._is_prev = False
                yield self._get_next(
                    ret_type or self._ret_type,
                    start_time=start_time,
                    update_current=update_current,
                )
                start_time = None
        except CroniterBadDateError:
            if self._max_years_btw_matches_explicitly_set:
                return
            else:
                raise

    def all_prev(self, ret_type=None, start_time=None, update_current=None):
        """Generate all previous dates."""
        try:
            while True:
                self._is_prev = True
                yield self._get_next(
                    ret_type or self._ret_type,
                    start_time=start_time,
                    update_current=update_current,
                )
                start_time = None
        except CroniterBadDateError:
            if self._max_years_btw_matches_explicitly_set:
                return
            else:
                raise

    def iter(self, *args, **kwargs):
        return self._is_prev and self.all_prev or self.all_next

    def __iter__(self):
        return self

    __next__ = next = _get_next

    def _calc(self, now, expanded, nth_weekday_of_month, is_prev):
        if is_prev:
            now = math.ceil(now)
            nearest_diff_method = self._get_prev_nearest_diff
            sign = -1
            offset = (len(expanded) > UNIX_CRON_LEN or now % 60 > 0) and 1 or 60
        else:
            now = math.floor(now)
            nearest_diff_method = self._get_next_nearest_diff
            sign = 1
            offset = (len(expanded) > UNIX_CRON_LEN) and 1 or 60

        dst = now = self._timestamp_to_datetime(now + sign * offset)

        month, year = dst.month, dst.year
        current_year = now.year
        DAYS = self.DAYS

        def proc_year(d):
            if len(expanded) == YEAR_CRON_LEN:
                try:
                    expanded[YEAR_FIELD].index("*")
                except ValueError:
                    # use None as range_val to indicate no loop
                    diff_year = nearest_diff_method(d.year, expanded[YEAR_FIELD], None)
                    if diff_year is None:
                        return None, d
                    elif diff_year != 0:
                        if is_prev:
                            d += relativedelta(
                                years=diff_year,
                                month=12,
                                day=31,
                                hour=23,
                                minute=59,
                                second=59,
                            )
                        else:
                            d += relativedelta(
                                years=diff_year,
                                month=1,
                                day=1,
                                hour=0,
                                minute=0,
                                second=0,
                            )
                        return True, d
            return False, d

        def proc_month(d):
            try:
                expanded[MONTH_FIELD].index("*")
            except ValueError:
                diff_month = nearest_diff_method(
                    d.month, expanded[MONTH_FIELD], self.MONTHS_IN_YEAR
                )
                days = DAYS[month - 1]
                if month == 2 and self.is_leap(year) is True:
                    days += 1

                reset_day = 1

                if diff_month is not None and diff_month != 0:
                    if is_prev:
                        d += relativedelta(months=diff_month)
                        reset_day = DAYS[d.month - 1]
                        if d.month == 2 and self.is_leap(d.year) is True:
                            reset_day += 1
                        d += relativedelta(day=reset_day, hour=23, minute=59, second=59)
                    else:
                        d += relativedelta(
                            months=diff_month, day=reset_day, hour=0, minute=0, second=0
                        )
                    return True, d
            return False, d

        def proc_day_of_month(d):
            try:
                expanded[DAY_FIELD].index("*")
            except ValueError:
                days = DAYS[month - 1]
                if month == 2 and self.is_leap(year) is True:
                    days += 1
                if "l" in expanded[DAY_FIELD] and days == d.day:
                    return False, d

                if is_prev:
                    days_in_prev_month = DAYS[(month - 2) % self.MONTHS_IN_YEAR]
                    diff_day = nearest_diff_method(
                        d.day, expanded[DAY_FIELD], days_in_prev_month
                    )
                else:
                    diff_day = nearest_diff_method(d.day, expanded[DAY_FIELD], days)

                if diff_day is not None and diff_day != 0:
                    if is_prev:
                        d += relativedelta(days=diff_day, hour=23, minute=59, second=59)
                    else:
                        d += relativedelta(days=diff_day, hour=0, minute=0, second=0)
                    return True, d
            return False, d

        def proc_day_of_week(d):
            try:
                expanded[DOW_FIELD].index("*")
            except ValueError:
                diff_day_of_week = nearest_diff_method(
                    d.isoweekday() % 7, expanded[DOW_FIELD], 7
                )
                if diff_day_of_week is not None and diff_day_of_week != 0:
                    if is_prev:
                        d += relativedelta(
                            days=diff_day_of_week, hour=23, minute=59, second=59
                        )
                    else:
                        d += relativedelta(
                            days=diff_day_of_week, hour=0, minute=0, second=0
                        )
                    return True, d
            return False, d

        def proc_day_of_week_nth(d):
            if "*" in nth_weekday_of_month:
                s = nth_weekday_of_month["*"]
                for i in range(0, 7):
                    if i in nth_weekday_of_month:
                        nth_weekday_of_month[i].update(s)
                    else:
                        nth_weekday_of_month[i] = s
                del nth_weekday_of_month["*"]

            candidates = []
            for wday, nth in nth_weekday_of_month.items():
                c = self._get_nth_weekday_of_month(d.year, d.month, wday)
                for n in nth:
                    if n == "l":
                        candidate = c[-1]
                    elif len(c) < n:
                        continue
                    else:
                        candidate = c[n - 1]
                    if (is_prev and candidate <= d.day) or (
                        not is_prev and d.day <= candidate
                    ):
                        candidates.append(candidate)

            if not candidates:
                if is_prev:
                    d += relativedelta(days=-d.day, hour=23, minute=59, second=59)
                else:
                    days = DAYS[month - 1]
                    if month == 2 and self.is_leap(year) is True:
                        days += 1
                    d += relativedelta(
                        days=(days - d.day + 1), hour=0, minute=0, second=0
                    )
                return True, d

            candidates.sort()
            diff_day = (candidates[-1] if is_prev else candidates[0]) - d.day
            if diff_day != 0:
                if is_prev:
                    d += relativedelta(days=diff_day, hour=23, minute=59, second=59)
                else:
                    d += relativedelta(days=diff_day, hour=0, minute=0, second=0)
                return True, d
            return False, d

        def proc_hour(d):
            try:
                expanded[HOUR_FIELD].index("*")
            except ValueError:
                diff_hour = nearest_diff_method(d.hour, expanded[HOUR_FIELD], 24)
                if diff_hour is not None and diff_hour != 0:
                    if is_prev:
                        d += relativedelta(hours=diff_hour, minute=59, second=59)
                    else:
                        d += relativedelta(hours=diff_hour, minute=0, second=0)
                    return True, d
            return False, d

        def proc_minute(d):
            try:
                expanded[MINUTE_FIELD].index("*")
            except ValueError:
                diff_min = nearest_diff_method(d.minute, expanded[MINUTE_FIELD], 60)
                if diff_min is not None and diff_min != 0:
                    if is_prev:
                        d += relativedelta(minutes=diff_min, second=59)
                    else:
                        d += relativedelta(minutes=diff_min, second=0)
                    return True, d
            return False, d

        def proc_second(d):
            if len(expanded) > UNIX_CRON_LEN:
                try:
                    expanded[SECOND_FIELD].index("*")
                except ValueError:
                    diff_sec = nearest_diff_method(d.second, expanded[SECOND_FIELD], 60)
                    if diff_sec is not None and diff_sec != 0:
                        d += relativedelta(seconds=diff_sec)
                        return True, d
            else:
                d += relativedelta(second=0)
            return False, d

        procs = [
            proc_year,
            proc_month,
            proc_day_of_month,
            (proc_day_of_week_nth if nth_weekday_of_month else proc_day_of_week),
            proc_hour,
            proc_minute,
            proc_second,
        ]

        while abs(year - current_year) <= self._max_years_between_matches:
            next = False
            stop = False
            for proc in procs:
                (changed, dst) = proc(dst)
                # `None` can be set mostly for year processing
                # so please see proc_year / _get_prev_nearest_diff / _get_next_nearest_diff
                if changed is None:
                    stop = True
                    break
                if changed:
                    month, year = dst.month, dst.year
                    next = True
                    break
            if stop:
                break
            if next:
                continue
            return self._datetime_to_timestamp(dst.replace(microsecond=0))

        if is_prev:
            raise CroniterBadDateError("failed to find prev date")
        raise CroniterBadDateError("failed to find next date")

    def _get_next_nearest(self, x, to_check):
        small = [item for item in to_check if item < x]
        large = [item for item in to_check if item >= x]
        large.extend(small)
        return large[0]

    def _get_prev_nearest(self, x, to_check):
        small = [item for item in to_check if item <= x]
        large = [item for item in to_check if item > x]
        small.reverse()
        large.reverse()
        small.extend(large)
        return small[0]

    def _get_next_nearest_diff(self, x, to_check, range_val):
        """
        Return diff between next and nearest.

        `range_val` is the range of a field.
        If no available time, we can move to next loop(like next month).
        `range_val` can also be set to `None` to indicate that there is no loop.
        ( Currently, should only used for `year` field )
        """
        for i, d in enumerate(to_check):
            if d == "l" and range_val is not None:
                # if 'l' then it is the last day of month
                # => its value of range_val
                d = range_val
            if d >= x:
                return d - x
        # When range_val is None and x not exists in to_check,
        # `None` will be returned to suggest no more available time
        if range_val is None:
            return None
        return to_check[0] - x + range_val

    def _get_prev_nearest_diff(self, x, to_check, range_val):
        """
        Return diff between previous and nearest.

        `range_val` is the range of a field.
        If no available time, we can move to previous loop(like previous month).
        Range_val can also be set to `None` to indicate that there is no loop.
        ( Currently should only used for `year` field )
        """
        candidates = to_check[:]
        candidates.reverse()
        for d in candidates:
            if d != "l" and d <= x:
                return d - x
        if "l" in candidates:
            return -x
        # When range_val is None and x not exists in to_check,
        # `None` will be returned to suggest no more available time
        if range_val is None:
            return None
        candidate = candidates[0]
        for c in candidates:
            # fixed: c < range_val
            # this code will reject all 31 day of month, 12 month, 59 second,
            # 23 hour and so on.
            # if candidates has just a element, this will not harmful.
            # but candidates have multiple elements, then values equal to
            # range_val will rejected.
            if c <= range_val:
                candidate = c
                break
        if candidate > range_val:
            # fix crontab "0 6 30 3 *" condidates only a element,
            # then get_prev error return 2021-03-02 06:00:00
            return -x
        return candidate - x - range_val

    @staticmethod
    def _get_nth_weekday_of_month(year, month, day_of_week):
        """
        Return list of days in nth-day-of-month order for a given year/month.

        The last weekday of the month is always [-1].
        """
        w = (day_of_week + 6) % 7
        c = calendar.Calendar(w).monthdayscalendar(year, month)
        if c[0][0] == 0:
            c.pop(0)
        return tuple(i[0] for i in c)

    def is_leap(self, year):
        if year % 400 == 0 or (year % 4 == 0 and year % 100 != 0):
            return True
        else:
            return False

    @classmethod
    def _expand(
        cls, expr_format, hash_id=None, second_at_beginning=False, from_timestamp=None
    ):
        # Split the expression in components, and normalize L -> l, MON -> mon,
        # etc. Keep expr_format untouched so we can use it in the exception
        # messages.
        expr_aliases = {
            "@midnight": ("0 0 * * *", "h h(0-2) * * * h"),
            "@hourly": ("0 * * * *", "h * * * * h"),
            "@daily": ("0 0 * * *", "h h * * * h"),
            "@weekly": ("0 0 * * 0", "h h * * h h"),
            "@monthly": ("0 0 1 * *", "h h h * * h"),
            "@yearly": ("0 0 1 1 *", "h h h h * h"),
            "@annually": ("0 0 1 1 *", "h h h h * h"),
        }

        efl = expr_format.lower()
        hash_id_expr = hash_id is not None and 1 or 0
        try:
            efl = expr_aliases[efl][hash_id_expr]
        except KeyError:
            pass

        expressions = efl.split()

        if len(expressions) not in VALID_LEN_EXPRESSION:
            raise CroniterBadCronError(cls.bad_length)

        if len(expressions) > UNIX_CRON_LEN and second_at_beginning:
            # move second to it's own(6th) field to process by same logical
            expressions.insert(SECOND_FIELD, expressions.pop(0))

        expanded = []
        nth_weekday_of_month = {}

        for i, expr in enumerate(expressions):
            for expanderid, expander in EXPANDERS.items():
                expr = expander(cls).expand(
                    efl, i, expr, hash_id=hash_id, from_timestamp=from_timestamp
                )

            if "?" in expr:
                if expr != "?":
                    raise CroniterBadCronError(
                        "[{0}] is not acceptable.  Question mark can not "
                        "used with other characters".format(expr_format)
                    )
                if i not in [DAY_FIELD, DOW_FIELD]:
                    raise CroniterBadCronError(
                        "[{0}] is not acceptable.  Question mark can only used "
                        "in day_of_month or day_of_week".format(expr_format)
                    )
                # currently just trade `?` as `*`
                expr = "*"

            e_list = expr.split(",")
            res = []

            while len(e_list) > 0:
                e = e_list.pop()
                nth = None

                if i == DOW_FIELD:
                    # Handle special case in the dow expression: 2#3, l3
                    special_dow_rem = special_dow_re.match(str(e))
                    if special_dow_rem:
                        g = special_dow_rem.groupdict()
                        he, last = g.get("he", ""), g.get("last", "")
                        if he:
                            e = he
                            try:
                                nth = int(last)
                                assert nth >= 1 and nth <= 5
                            except (KeyError, ValueError, AssertionError):
                                raise CroniterBadCronError(
                                    "[{0}] is not acceptable.  Invalid day_of_week "
                                    "value: '{1}'".format(expr_format, nth)
                                )
                        elif last:
                            e = last
                            nth = g["pre"]  # 'l'

                # Before matching step_search_re, normalize "*" to "{min}-{max}".
                # Example: in the minute field, "*/5" normalizes to "0-59/5"
                t = re.sub(
                    r"^\*(\/.+)$",
                    r"%d-%d\1" % (cls.RANGES[i][0], cls.RANGES[i][1]),
                    str(e),
                )
                m = step_search_re.search(t)

                if not m:
                    # Before matching step_search_re,
                    # normalize "{start}/{step}" to "{start}-{max}/{step}".
                    # Example: in the minute field, "10/5" normalizes to "10-59/5"
                    t = re.sub(
                        r"^(.+)\/(.+)$", r"\1-%d/\2" % (cls.RANGES[i][1]), str(e)
                    )
                    m = step_search_re.search(t)

                if m:
                    # early abort if low/high are out of bounds

                    (low, high, step) = m.group(1), m.group(2), m.group(4) or 1
                    if i == DAY_FIELD and high == "l":
                        high = "31"

                    if not only_int_re.search(low):
                        low = "{0}".format(cls._alphaconv(i, low, expressions))

                    if not only_int_re.search(high):
                        high = "{0}".format(cls._alphaconv(i, high, expressions))

                    if (
                        not low
                        or not high
                        or int(low) > int(high)
                        or not only_int_re.search(str(step))
                    ):
                        if i == DOW_FIELD and high == "0":
                            # handle -Sun notation -> 7
                            high = "7"
                        else:
                            raise CroniterBadCronError(
                                "[{0}] is not acceptable".format(expr_format)
                            )

                    low, high, step = map(int, [low, high, step])
                    if max(low, high) > max(cls.RANGES[i][0], cls.RANGES[i][1]):
                        raise CroniterBadCronError(
                            "{0} is out of bands".format(expr_format)
                        )

                    if from_timestamp:
                        low = cls._get_low_from_current_date_number(
                            i, int(step), int(from_timestamp)
                        )

                    try:
                        rng = range(low, high + 1, step)
                    except ValueError as exc:
                        raise CroniterBadCronError("invalid range: {0}".format(exc))
                    e_list += (
                        ["{0}#{1}".format(item, nth) for item in rng]
                        if i == DOW_FIELD and nth and nth != "l"
                        else rng
                    )
                else:
                    if t.startswith("-"):
                        raise CroniterBadCronError(
                            (
                                "[{0}] is not acceptable,"
                                "negative numbers not allowed"
                            ).format(expr_format)
                        )
                    if not star_or_int_re.search(t):
                        t = cls._alphaconv(i, t, expressions)

                    try:
                        t = int(t)
                    except ValueError:
                        pass

                    if t in cls.LOWMAP[i] and not (
                        # do not support 0 as a month either for classical 5 fields cron,
                        # 6fields second repeat form or 7 fields year form
                        # but still let conversion happen if day field is shifted
                        (
                            i in [DAY_FIELD, MONTH_FIELD]
                            and len(expressions) == UNIX_CRON_LEN
                        )
                        or (
                            i in [MONTH_FIELD, DOW_FIELD]
                            and len(expressions) == SECOND_CRON_LEN
                        )
                        or (
                            i in [DAY_FIELD, MONTH_FIELD, DOW_FIELD]
                            and len(expressions) == YEAR_CRON_LEN
                        )
                    ):
                        t = cls.LOWMAP[i][t]

                    if t not in ["*", "l"] and (
                        int(t) < cls.RANGES[i][0] or int(t) > cls.RANGES[i][1]
                    ):
                        raise CroniterBadCronError(
                            "[{0}] is not acceptable, out of range".format(expr_format)
                        )

                    res.append(t)

                    if i == DOW_FIELD and nth:
                        if t not in nth_weekday_of_month:
                            nth_weekday_of_month[t] = set()
                        nth_weekday_of_month[t].add(nth)

            res = set(res)
            res = sorted(
                res, key=lambda i: "{:02}".format(i) if isinstance(i, int) else i
            )
            if len(res) == cls.LEN_MEANS_ALL[i]:
                # Make sure the wildcard is used in the correct way (avoid over-optimization)
                if (i == DAY_FIELD and "*" not in expressions[DOW_FIELD]) or (
                    i == DOW_FIELD and "*" not in expressions[DAY_FIELD]
                ):
                    pass
                else:
                    res = ["*"]

            expanded.append(["*"] if (len(res) == 1 and res[0] == "*") else res)

        # Check to make sure the dow combo in use is supported
        if nth_weekday_of_month:
            dow_expanded_set = set(expanded[DOW_FIELD])
            dow_expanded_set = dow_expanded_set.difference(nth_weekday_of_month.keys())
            dow_expanded_set.discard("*")
            # Skip: if it's all weeks instead of wildcard
            if (
                dow_expanded_set
                and len(set(expanded[DOW_FIELD])) != cls.LEN_MEANS_ALL[DOW_FIELD]
            ):
                raise CroniterUnsupportedSyntaxError(
                    "day-of-week field does not support mixing literal values and nth day of week syntax.  "
                    "Cron: '{}'    dow={} vs nth={}".format(
                        expr_format, dow_expanded_set, nth_weekday_of_month
                    )
                )

        EXPRESSIONS[(expr_format, hash_id, second_at_beginning)] = expressions
        return expanded, nth_weekday_of_month

    @classmethod
    def expand(
        cls, expr_format, hash_id=None, second_at_beginning=False, from_timestamp=None
    ):
        """Shallow non Croniter ValueError inside a nice CroniterBadCronError."""
        try:
            return cls._expand(
                expr_format,
                hash_id=hash_id,
                second_at_beginning=second_at_beginning,
                from_timestamp=from_timestamp,
            )
        except (ValueError,) as exc:
            error_type, error_instance, traceback = sys.exc_info()
            if isinstance(exc, CroniterError):
                raise
            if int(sys.version[0]) >= 3:
                trace = _traceback.format_exc()
                globs, locs = _get_caller_globals_and_locals()
                raise CroniterBadCronError(trace)
            else:
                raise CroniterBadCronError("{0}".format(exc))

    @classmethod
    def _get_low_from_current_date_number(cls, i, step, from_timestamp):
        dt = datetime.datetime.fromtimestamp(from_timestamp, tz=datetime.timezone.utc)
        if i == MINUTE_FIELD:
            return dt.minute % step
        if i == HOUR_FIELD:
            return dt.hour % step
        if i == DAY_FIELD:
            return ((dt.day - 1) % step) + 1
        if i == MONTH_FIELD:
            return dt.month % step
        if i == DOW_FIELD:
            return (dt.weekday() + 1) % step

        raise ValueError("Can't get current date number for index larger than 4")

    @classmethod
    def is_valid(
        cls, expression, hash_id=None, encoding="UTF-8", second_at_beginning=False
    ):
        if hash_id:
            if not isinstance(hash_id, (bytes, str)):
                raise TypeError("hash_id must be bytes or UTF-8 string")
            if not isinstance(hash_id, bytes):
                hash_id = hash_id.encode(encoding)
        try:
            cls.expand(
                expression, hash_id=hash_id, second_at_beginning=second_at_beginning
            )
        except CroniterError:
            return False
        else:
            return True

    @classmethod
    def match(cls, cron_expression, testdate, day_or=True, second_at_beginning=False):
        return cls.match_range(
            cron_expression, testdate, testdate, day_or, second_at_beginning
        )

    @classmethod
    def match_range(
        cls,
        cron_expression,
        from_datetime,
        to_datetime,
        day_or=True,
        second_at_beginning=False,
    ):
        cron = cls(
            cron_expression,
            to_datetime,
            ret_type=datetime.datetime,
            day_or=day_or,
            second_at_beginning=second_at_beginning,
        )
        tdp = cron.get_current(datetime.datetime)
        if not tdp.microsecond:
            tdp += relativedelta(microseconds=1)
        cron.set_current(tdp, force=True)
        try:
            tdt = cron.get_prev()
        except CroniterBadDateError:
            return False
        precision_in_seconds = 1 if len(cron.expanded) > UNIX_CRON_LEN else 60
        duration_in_second = (
            to_datetime - from_datetime
        ).total_seconds() + precision_in_seconds
        return (max(tdp, tdt) - min(tdp, tdt)).total_seconds() < duration_in_second


def croniter_range(
    start,
    stop,
    expr_format,
    ret_type=None,
    day_or=True,
    exclude_ends=False,
    _croniter=None,
    second_at_beginning=False,
    expand_from_start_time=False,
):
    """
    Generate all times from start to stop matching the given cron expression.

    If the cron expression matches either 'start' and/or 'stop', those times will be returned as
    well unless 'exclude_ends=True' is passed.

    You can think of this function as sibling to the builtin range function for datetime objects.
    Like range(start,stop,step), except that here 'step' is a cron expression.
    """
    _croniter = _croniter or croniter
    auto_rt = datetime.datetime
    # type is used in first if branch for perfs reasons
    if type(start) is not type(stop) and not (
        isinstance(start, type(stop)) or isinstance(stop, type(start))
    ):
        raise CroniterBadTypeRangeError(
            "The start and stop must be same type.  {0} != {1}".format(
                type(start), type(stop)
            )
        )
    if isinstance(start, (float, int)):
        start, stop = (
            datetime.datetime.fromtimestamp(t, tzutc()).replace(tzinfo=None)
            for t in (start, stop)
        )
        auto_rt = float
    if ret_type is None:
        ret_type = auto_rt
    if not exclude_ends:
        ms1 = relativedelta(microseconds=1)
        if start < stop:  # Forward (normal) time order
            start -= ms1
            stop += ms1
        else:  # Reverse time order
            start += ms1
            stop -= ms1
    year_span = math.floor(abs(stop.year - start.year)) + 1
    ic = _croniter(
        expr_format,
        start,
        ret_type=datetime.datetime,
        day_or=day_or,
        max_years_between_matches=year_span,
        second_at_beginning=second_at_beginning,
        expand_from_start_time=expand_from_start_time,
    )
    # define a continue (cont) condition function and step function for the main while loop
    if start < stop:  # Forward

        def cont(v):
            return v < stop

        step = ic.get_next
    else:  # Reverse

        def cont(v):
            return v > stop

        step = ic.get_prev
    try:
        dt = step()
        while cont(dt):
            if ret_type is float:
                yield ic.get_current(float)
            else:
                yield dt
            dt = step()
    except CroniterBadDateError:
        # Stop iteration when this exception is raised; no match found within the given year range
        return


class HashExpander:

    def __init__(self, cronit):
        self.cron = cronit

    def do(self, idx, hash_type="h", hash_id=None, range_end=None, range_begin=None):
        """Return a hashed/random integer given range/hash information."""
        if range_end is None:
            range_end = self.cron.RANGES[idx][1]
        if range_begin is None:
            range_begin = self.cron.RANGES[idx][0]
        if hash_type == "r":
            crc = random.randint(0, 0xFFFFFFFF)
        else:
            crc = binascii.crc32(hash_id) & 0xFFFFFFFF
        return ((crc >> idx) % (range_end - range_begin + 1)) + range_begin

    def match(self, efl, idx, expr, hash_id=None, **kw):
        return hash_expression_re.match(expr)

    def expand(self, efl, idx, expr, hash_id=None, match="", **kw):
        """Expand a hashed/random expression to its normal representation."""
        if match == "":
            match = self.match(efl, idx, expr, hash_id, **kw)
        if not match:
            return expr
        m = match.groupdict()

        if m["hash_type"] == "h" and hash_id is None:
            raise CroniterBadCronError("Hashed definitions must include hash_id")

        if m["range_begin"] and m["range_end"]:
            if int(m["range_begin"]) >= int(m["range_end"]):
                raise CroniterBadCronError("Range end must be greater than range begin")

        if m["range_begin"] and m["range_end"] and m["divisor"]:
            # Example: H(30-59)/10 -> 34-59/10 (i.e. 34,44,54)
            if int(m["divisor"]) == 0:
                raise CroniterBadCronError("Bad expression: {0}".format(expr))

            return "{0}-{1}/{2}".format(
                self.do(
                    idx,
                    hash_type=m["hash_type"],
                    hash_id=hash_id,
                    range_begin=int(m["range_begin"]),
                    range_end=int(m["divisor"]) - 1 + int(m["range_begin"]),
                ),
                int(m["range_end"]),
                int(m["divisor"]),
            )
        elif m["range_begin"] and m["range_end"]:
            # Example: H(0-29) -> 12
            return str(
                self.do(
                    idx,
                    hash_type=m["hash_type"],
                    hash_id=hash_id,
                    range_end=int(m["range_end"]),
                    range_begin=int(m["range_begin"]),
                )
            )
        elif m["divisor"]:
            # Example: H/15 -> 7-59/15 (i.e. 7,22,37,52)
            if int(m["divisor"]) == 0:
                raise CroniterBadCronError("Bad expression: {0}".format(expr))

            return "{0}-{1}/{2}".format(
                self.do(
                    idx,
                    hash_type=m["hash_type"],
                    hash_id=hash_id,
                    range_begin=self.cron.RANGES[idx][0],
                    range_end=int(m["divisor"]) - 1 + self.cron.RANGES[idx][0],
                ),
                self.cron.RANGES[idx][1],
                int(m["divisor"]),
            )
        else:
            # Example: H -> 32
            return str(
                self.do(
                    idx,
                    hash_type=m["hash_type"],
                    hash_id=hash_id,
                )
            )


EXPANDERS = OrderedDict(
    [
        ("hash", HashExpander),
    ]
)



"""Errors thrown by DBOS."""

from enum import Enum
from typing import Optional


class DBOSException(Exception):
    """
    Base class of DBOS Exceptions.

    Attributes:
        message(str): The error message string
        dbos_error_code(DBOSErrorCode): The error code, from the `DBOSErrorCode` enum

    """

    def __init__(self, message: str, dbos_error_code: Optional[int] = None):
        self.message = message
        self.dbos_error_code = dbos_error_code
        self.status_code: Optional[int] = None
        super().__init__(self.message)

    def __str__(self) -> str:
        if self.dbos_error_code:
            return f"DBOS Error {self.dbos_error_code}: {self.message}"
        return f"DBOS Error: {self.message}"


class DBOSErrorCode(Enum):
    ConflictingIDError = 1
    RecoveryError = 2
    InitializationError = 3
    WorkflowFunctionNotFound = 4
    NonExistentWorkflowError = 5
    DeadLetterQueueError = 6
    MaxStepRetriesExceeded = 7
    NotAuthorized = 8
    ConflictingWorkflowError = 9
    WorkflowCancelled = 10
    ConflictingRegistrationError = 25


class DBOSWorkflowConflictIDError(DBOSException):
    """Exception raised when a workflow database record already exists."""

    def __init__(self, workflow_id: str):
        super().__init__(
            f"Conflicting workflow ID {workflow_id}",
            dbos_error_code=DBOSErrorCode.ConflictingIDError.value,
        )


class DBOSConflictingWorkflowError(DBOSException):
    """Exception raised different workflows started with the same workflow ID."""

    def __init__(self, workflow_id: str, message: Optional[str] = None):
        super().__init__(
            f"Conflicting workflow invocation with the same ID ({workflow_id}): {message}",
            dbos_error_code=DBOSErrorCode.ConflictingWorkflowError.value,
        )


class DBOSRecoveryError(DBOSException):
    """Exception raised when a workflow recovery fails."""

    def __init__(self, workflow_id: str, message: Optional[str] = None):
        super().__init__(
            f"Recovery error for workflow ID {workflow_id}: {message}",
            dbos_error_code=DBOSErrorCode.RecoveryError.value,
        )


class DBOSInitializationError(DBOSException):
    """Exception raised when DBOS initialization did not complete."""

    def __init__(self, message: str):
        super().__init__(
            f"Error initializing DBOS Transact: {message}",
            DBOSErrorCode.InitializationError.value,
        )


class DBOSWorkflowFunctionNotFoundError(DBOSException):
    """Exception raised when the database refers to a workflow function that is not registered in the codebase."""

    def __init__(self, workflow_id: str, message: Optional[str] = None):
        super().__init__(
            f"Workflow function not found for workflow ID {workflow_id}: {message}",
            dbos_error_code=DBOSErrorCode.WorkflowFunctionNotFound.value,
        )


class DBOSNonExistentWorkflowError(DBOSException):
    """Exception raised when a workflow database record does not exist for a given ID."""

    def __init__(self, destination_id: str):
        super().__init__(
            f"Sent to non-existent destination workflow ID: {destination_id}",
            dbos_error_code=DBOSErrorCode.NonExistentWorkflowError.value,
        )


class DBOSDeadLetterQueueError(DBOSException):
    """Exception raised when a workflow database record does not exist for a given ID."""

    def __init__(self, wf_id: str, max_retries: int):
        super().__init__(
            f"Workflow {wf_id} has been moved to the dead-letter queue after exceeding the maximum of ${max_retries} retries",
            dbos_error_code=DBOSErrorCode.DeadLetterQueueError.value,
        )


class DBOSNotAuthorizedError(DBOSException):
    """Exception raised by DBOS role-based security when the user is not authorized to access a function."""

    def __init__(self, msg: str):
        super().__init__(
            msg,
            dbos_error_code=DBOSErrorCode.NotAuthorized.value,
        )
        self.status_code = 403


class DBOSMaxStepRetriesExceeded(DBOSException):
    """Exception raised when a step was retried the maximimum number of times without success."""

    def __init__(self) -> None:
        super().__init__(
            "Step reached maximum retries.",
            dbos_error_code=DBOSErrorCode.MaxStepRetriesExceeded.value,
        )


class DBOSWorkflowCancelledError(DBOSException):
    """Exception raised when the workflow has already been cancelled."""

    def __init__(self, msg: str) -> None:
        super().__init__(
            msg,
            dbos_error_code=DBOSErrorCode.WorkflowCancelled.value,
        )


class DBOSConflictingRegistrationError(DBOSException):
    """Exception raised when conflicting decorators are applied to the same function."""

    def __init__(self, name: str) -> None:
        super().__init__(
            f"Operation (Name: {name}) is already registered with a conflicting function type",
            dbos_error_code=DBOSErrorCode.ConflictingRegistrationError.value,
        )



"""
Adjust workflow queue to add columns for rate limiter.

Revision ID: d76646551a6b
Revises: 50f3227f0b4b
Create Date: 2024-09-25 14:48:10.218015

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "d76646551a6b"
down_revision: Union[str, None] = "50f3227f0b4b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column(
        "job_queue",
        sa.Column(
            "started_at_epoch_ms",
            sa.BigInteger(),
        ),
        schema="dbos",
    )
    op.add_column(
        "job_queue",
        sa.Column(
            "completed_at_epoch_ms",
            sa.BigInteger(),
        ),
        schema="dbos",
    )


def downgrade() -> None:
    op.drop_column("job_queue", "started_at_epoch_ms", schema="dbos")
    op.drop_column("job_queue", "completed_at_epoch_ms", schema="dbos")



import json
import os
import re
from importlib import resources
from typing import Any, Dict, List, Optional, TypedDict, cast

import yaml
from jsonschema import ValidationError, validate
from rich import print
from sqlalchemy import URL

from ._db_wizard import db_wizard, load_db_connection
from ._error import DBOSInitializationError
from ._logger import config_logger, dbos_logger, init_logger

DBOS_CONFIG_PATH = "dbos-config.yaml"


class RuntimeConfig(TypedDict, total=False):
    start: List[str]
    setup: Optional[List[str]]
    admin_port: Optional[int]


class DatabaseConfig(TypedDict, total=False):
    hostname: str
    port: int
    username: str
    password: str
    connectionTimeoutMillis: Optional[int]
    app_db_name: str
    sys_db_name: Optional[str]
    ssl: Optional[bool]
    ssl_ca: Optional[str]
    local_suffix: Optional[bool]
    app_db_client: Optional[str]
    migrate: Optional[List[str]]
    rollback: Optional[List[str]]


class OTLPExporterConfig(TypedDict, total=False):
    logsEndpoint: Optional[str]
    tracesEndpoint: Optional[str]


class LoggerConfig(TypedDict, total=False):
    logLevel: Optional[str]


class TelemetryConfig(TypedDict, total=False):
    logs: Optional[LoggerConfig]
    OTLPExporter: Optional[OTLPExporterConfig]


class ConfigFile(TypedDict, total=False):
    """
    Data structure containing the DBOS Configuration.

    This configuration data is typically loaded from `dbos-config.yaml`.
    See `https://docs.dbos.dev/api-reference/configuration`_

    Attributes:
        name (str): Application name
        language (str): The app language (probably `python`)
        runtimeConfig (RuntimeConfig): Configuration for request serving
        database (DatabaseConfig): Configuration for the application and system databases
        telemetry (TelemetryConfig): Configuration for tracing / logging
        env (Dict[str,str]): Environment varialbes
        application (Dict[str, Any]): Application-specific configuration section

    """

    name: str
    language: str
    runtimeConfig: RuntimeConfig
    database: DatabaseConfig
    telemetry: Optional[TelemetryConfig]
    env: Dict[str, str]
    application: Dict[str, Any]


def _substitute_env_vars(content: str) -> str:
    regex = r"\$\{([^}]+)\}"  # Regex to match ${VAR_NAME} style placeholders

    def replace_func(match: re.Match[str]) -> str:
        var_name = match.group(1)
        value = os.environ.get(
            var_name, ""
        )  # If the env variable is not set, return an empty string
        if value == "":
            dbos_logger.warning(
                f"Variable {var_name} would be substituted from the process environment into dbos-config.yaml, but is not defined"
            )
        return value

    return re.sub(regex, replace_func, content)


def get_dbos_database_url(config_file_path: str = DBOS_CONFIG_PATH) -> str:
    """
    Retrieve application database URL from configuration `.yaml` file.

    Loads the DBOS `ConfigFile` from the specified path (typically `dbos-config.yaml`),
        and returns the database URL for the application database.

    Args:
        config_file_path (str): The path to the yaml configuration file.

    Returns:
        str: Database URL for the application database

    """
    dbos_config = load_config(config_file_path)
    db_url = URL.create(
        "postgresql+psycopg",
        username=dbos_config["database"]["username"],
        password=dbos_config["database"]["password"],
        host=dbos_config["database"]["hostname"],
        port=dbos_config["database"]["port"],
        database=dbos_config["database"]["app_db_name"],
    )
    return db_url.render_as_string(hide_password=False)


def load_config(
    config_file_path: str = DBOS_CONFIG_PATH,
    *,
    use_db_wizard: bool = True,
    silent: bool = False,
) -> ConfigFile:
    """
    Load the DBOS `ConfigFile` from the specified path (typically `dbos-config.yaml`).

    The configuration is also validated against the configuration file schema.

    Args:
        config_file_path (str): The path to the yaml configuration file.

    Returns:
        ConfigFile: The loaded configuration

    """

    init_logger()

    with open(config_file_path, "r") as file:
        content = file.read()
        substituted_content = _substitute_env_vars(content)
        data = yaml.safe_load(substituted_content)

    # Load the JSON schema relative to the package root
    schema_file = resources.files("dbos").joinpath("dbos-config.schema.json")
    with schema_file.open("r") as f:
        schema = json.load(f)

    # Validate the data against the schema
    try:
        validate(instance=data, schema=schema)
    except ValidationError as e:
        raise DBOSInitializationError(f"Validation error: {e}")

    if "database" not in data:
        data["database"] = {}

    if "name" not in data:
        raise DBOSInitializationError(
            f"dbos-config.yaml must specify an application name"
        )

    if "language" not in data:
        raise DBOSInitializationError(
            f"dbos-config.yaml must specify the application language is Python"
        )

    if data["language"] != "python":
        raise DBOSInitializationError(
            f'dbos-config.yaml specifies invalid language { data["language"] }'
        )

    if "runtimeConfig" not in data or "start" not in data["runtimeConfig"]:
        raise DBOSInitializationError(f"dbos-config.yaml must specify a start command")

    if not _is_valid_app_name(data["name"]):
        raise DBOSInitializationError(
            f'Invalid app name {data["name"]}.  App names must be between 3 and 30 characters long and contain only lowercase letters, numbers, dashes, and underscores.'
        )

    if "app_db_name" not in data["database"]:
        data["database"]["app_db_name"] = _app_name_to_db_name(data["name"])

    # Load the DB connection file. Use its values for missing fields from dbos-config.yaml. Use defaults otherwise.
    data = cast(ConfigFile, data)
    db_connection = load_db_connection()
    if not silent:
        if os.getenv("DBOS_DBHOST"):
            print(
                "[bold blue]Loading database connection parameters from debug environment variables[/bold blue]"
            )
        elif data["database"].get("hostname"):
            print(
                "[bold blue]Loading database connection parameters from dbos-config.yaml[/bold blue]"
            )
        elif db_connection.get("hostname"):
            print(
                "[bold blue]Loading database connection parameters from .dbos/db_connection[/bold blue]"
            )
        else:
            print(
                "[bold blue]Using default database connection parameters (localhost)[/bold blue]"
            )

    dbos_dbport: Optional[int] = None
    dbport_env = os.getenv("DBOS_DBPORT")
    if dbport_env:
        try:
            dbos_dbport = int(dbport_env)
        except ValueError:
            pass
    dbos_dblocalsuffix: Optional[bool] = None
    dblocalsuffix_env = os.getenv("DBOS_DBLOCALSUFFIX")
    if dblocalsuffix_env:
        try:
            dbos_dblocalsuffix = dblocalsuffix_env.casefold() == "true".casefold()
        except ValueError:
            pass

    data["database"]["hostname"] = (
        os.getenv("DBOS_DBHOST")
        or data["database"].get("hostname")
        or db_connection.get("hostname")
        or "localhost"
    )

    data["database"]["port"] = (
        dbos_dbport or data["database"].get("port") or db_connection.get("port") or 5432
    )
    data["database"]["username"] = (
        os.getenv("DBOS_DBUSER")
        or data["database"].get("username")
        or db_connection.get("username")
        or "postgres"
    )
    data["database"]["password"] = (
        os.getenv("DBOS_DBPASSWORD")
        or data["database"].get("password")
        or db_connection.get("password")
        or os.environ.get("PGPASSWORD")
        or "dbos"
    )

    local_suffix = False
    dbcon_local_suffix = db_connection.get("local_suffix")
    if dbcon_local_suffix is not None:
        local_suffix = dbcon_local_suffix
    if data["database"].get("local_suffix") is not None:
        local_suffix = data["database"].get("local_suffix")
    if dbos_dblocalsuffix is not None:
        local_suffix = dbos_dblocalsuffix
    data["database"]["local_suffix"] = local_suffix

    # Configure the DBOS logger
    config_logger(data)

    # Check the connectivity to the database and make sure it's properly configured
    # Note, never use db wizard if the DBOS is running in debug mode (i.e. DBOS_DEBUG_WORKFLOW_ID env var is set)
    debugWorkflowId = os.getenv("DBOS_DEBUG_WORKFLOW_ID")
    if use_db_wizard and debugWorkflowId is None:
        data = db_wizard(data, config_file_path)

    if "local_suffix" in data["database"] and data["database"]["local_suffix"]:
        data["database"]["app_db_name"] = f"{data['database']['app_db_name']}_local"

    # Return data as ConfigFile type
    return data  # type: ignore


def _is_valid_app_name(name: str) -> bool:
    name_len = len(name)
    if name_len < 3 or name_len > 30:
        return False
    match = re.match("^[a-z0-9-_]+$", name)
    return True if match != None else False


def _app_name_to_db_name(app_name: str) -> str:
    name = app_name.replace("-", "_")
    return name if not name[0].isdigit() else f"_{name}"


def set_env_vars(config: ConfigFile) -> None:
    for env, value in config.get("env", {}).items():
        if value is not None:
            os.environ[env] = str(value)



import asyncio
import time
import uuid
from typing import Optional

import pytest
import sqlalchemy as sa

# Public API
from dbos import DBOS, SetWorkflowID
from dbos._dbos_config import ConfigFile
from dbos._error import DBOSException


@pytest.mark.asyncio
async def test_async_workflow(dbos: DBOS) -> None:
    txn_counter: int = 0
    wf_counter: int = 0
    step_counter: int = 0

    @DBOS.workflow()
    async def test_workflow(var1: str, var2: str) -> str:
        nonlocal wf_counter
        wf_counter += 1
        res1 = test_transaction(var1)
        res2 = test_step(var2)
        DBOS.logger.info("I'm test_workflow")
        return res1 + res2

    @DBOS.step()
    def test_step(var: str) -> str:
        nonlocal step_counter
        step_counter += 1
        DBOS.logger.info("I'm test_step")
        return var + f"step{step_counter}"

    @DBOS.transaction(isolation_level="REPEATABLE READ")
    def test_transaction(var: str) -> str:
        rows = (DBOS.sql_session.execute(sa.text("SELECT 1"))).fetchall()
        nonlocal txn_counter
        txn_counter += 1
        DBOS.logger.info("I'm test_transaction")
        return var + f"txn{txn_counter}{rows[0][0]}"

    wfuuid = f"test_async_workflow-{time.time_ns()}"
    with SetWorkflowID(wfuuid):
        result = await test_workflow("alice", "bob")
        assert result == "alicetxn11bobstep1"
    dbos._sys_db.wait_for_buffer_flush()

    with SetWorkflowID(wfuuid):
        result = await test_workflow("alice", "bob")
        assert result == "alicetxn11bobstep1"

    assert wf_counter == 2
    assert step_counter == 1
    assert txn_counter == 1


@pytest.mark.asyncio
async def test_async_step(dbos: DBOS) -> None:
    txn_counter: int = 0
    wf_counter: int = 0
    step_counter: int = 0

    @DBOS.workflow()
    async def test_workflow(var1: str, var2: str) -> str:
        nonlocal wf_counter
        wf_counter += 1
        res1 = test_transaction(var1)
        res2 = await test_step(var2)
        DBOS.logger.info("I'm test_workflow")
        return res1 + res2

    @DBOS.step()
    async def test_step(var: str) -> str:
        await asyncio.sleep(0.1)
        nonlocal step_counter
        step_counter += 1
        DBOS.logger.info("I'm test_step")
        return var + f"step{step_counter}"

    @DBOS.transaction(isolation_level="REPEATABLE READ")
    def test_transaction(var: str) -> str:
        rows = (DBOS.sql_session.execute(sa.text("SELECT 1"))).fetchall()
        nonlocal txn_counter
        txn_counter += 1
        DBOS.logger.info("I'm test_transaction")
        return var + f"txn{txn_counter}{rows[0][0]}"

    wfuuid = f"test_async_step-{time.time_ns()}"
    with SetWorkflowID(wfuuid):
        result = await test_workflow("alice", "bob")
        assert result == "alicetxn11bobstep1"
    dbos._sys_db.wait_for_buffer_flush()

    with SetWorkflowID(wfuuid):
        result = await test_workflow("alice", "bob")
        assert result == "alicetxn11bobstep1"

    assert wf_counter == 2
    assert step_counter == 1
    assert txn_counter == 1


@pytest.mark.asyncio
async def test_send_recv_async(dbos: DBOS) -> None:
    send_counter: int = 0
    recv_counter: int = 0

    @DBOS.workflow()
    async def test_send_workflow(dest_uuid: str, topic: str) -> str:
        await dbos.send_async(dest_uuid, "test1")
        await dbos.send_async(dest_uuid, "test2", topic=topic)
        await dbos.send_async(dest_uuid, "test3")
        nonlocal send_counter
        send_counter += 1
        return dest_uuid

    @DBOS.workflow()
    async def test_recv_workflow(topic: str) -> str:
        msg1 = await dbos.recv_async(topic, timeout_seconds=10)
        msg2 = await dbos.recv_async(timeout_seconds=10)
        msg3 = await dbos.recv_async(timeout_seconds=10)
        nonlocal recv_counter
        recv_counter += 1
        return "-".join([str(msg1), str(msg2), str(msg3)])

    @DBOS.workflow()
    async def test_recv_timeout(timeout_seconds: float) -> None:
        msg = await dbos.recv_async(timeout_seconds=timeout_seconds)
        assert msg is None

    @DBOS.workflow()
    async def test_send_none(dest_uuid: str) -> None:
        await dbos.send_async(dest_uuid, None)

    dest_uuid = str(uuid.uuid4())

    # Send to non-existent uuid should fail
    with pytest.raises(Exception) as exc_info:
        await test_send_workflow(dest_uuid, "testtopic")
    assert f"Sent to non-existent destination workflow ID: {dest_uuid}" in str(
        exc_info.value
    )

    with SetWorkflowID(dest_uuid):
        handle = dbos.start_workflow(test_recv_workflow, "testtopic")
        assert handle.get_workflow_id() == dest_uuid

    send_uuid = str(uuid.uuid4())
    with SetWorkflowID(send_uuid):
        res = await test_send_workflow(handle.get_workflow_id(), "testtopic")
        assert res == dest_uuid
    begin_time = time.time()
    assert handle.get_result() == "test2-test1-test3"
    duration = time.time() - begin_time
    assert duration < 3.0  # Shouldn't take more than 3 seconds to run

    # Test send 'None'
    none_uuid = str(uuid.uuid4())
    none_handle = None
    with SetWorkflowID(none_uuid):
        none_handle = dbos.start_workflow(test_recv_timeout, 10.0)
    await test_send_none(none_uuid)
    begin_time = time.time()
    assert none_handle.get_result() is None
    duration = time.time() - begin_time
    assert duration < 1.0  # None is from the received message, not from the timeout.

    timeout_uuid = str(uuid.uuid4())
    with SetWorkflowID(timeout_uuid):
        begin_time = time.time()
        await test_recv_timeout(1.0)
        duration = time.time() - begin_time
        assert duration > 0.7

    # Test OAOO
    with SetWorkflowID(send_uuid):
        res = await test_send_workflow(handle.get_workflow_id(), "testtopic")
        assert res == dest_uuid
        assert send_counter == 2

    with SetWorkflowID(dest_uuid):
        begin_time = time.time()
        res = await test_recv_workflow("testtopic")
        duration = time.time() - begin_time
        assert duration < 3.0
        assert res == "test2-test1-test3"
        assert recv_counter == 2

    with SetWorkflowID(timeout_uuid):
        begin_time = time.time()
        await test_recv_timeout(1.0)
        duration = time.time() - begin_time
        assert duration < 0.3

    # Test recv outside of a workflow
    with pytest.raises(Exception) as exc_info:
        await dbos.recv_async("test1")
    assert "recv() must be called from within a workflow" in str(exc_info.value)


@pytest.mark.asyncio
async def test_set_get_events(dbos: DBOS) -> None:
    @DBOS.workflow()
    async def test_setevent_workflow() -> None:
        await dbos.set_event_async("key1", "value1")
        await dbos.set_event_async("key2", "value2")
        await dbos.set_event_async("key3", None)

    @DBOS.workflow()
    async def test_getevent_workflow(
        target_uuid: str, key: str, timeout_seconds: float = 10
    ) -> Optional[str]:
        msg = await dbos.get_event_async(target_uuid, key, timeout_seconds)
        return str(msg) if msg is not None else None

    wfuuid = str(uuid.uuid4())
    with SetWorkflowID(wfuuid):
        await test_setevent_workflow()
    with SetWorkflowID(wfuuid):
        await test_setevent_workflow()

    value1 = await test_getevent_workflow(wfuuid, "key1")
    assert value1 == "value1"

    value2 = await test_getevent_workflow(wfuuid, "key2")
    assert value2 == "value2"

    # Run getEvent outside of a workflow
    value1 = await dbos.get_event_async(wfuuid, "key1")
    assert value1 == "value1"

    value2 = await dbos.get_event_async(wfuuid, "key2")
    assert value2 == "value2"

    begin_time = time.time()
    value3 = await test_getevent_workflow(wfuuid, "key3")
    assert value3 is None
    duration = time.time() - begin_time
    assert duration < 1  # None is from the event not from the timeout

    # Test OAOO
    timeout_uuid = str(uuid.uuid4())
    with SetWorkflowID(timeout_uuid):
        begin_time = time.time()
        res = await test_getevent_workflow("non-existent-uuid", "key1", 1.0)
        duration = time.time() - begin_time
        assert duration > 0.7
        assert res is None

    with SetWorkflowID(timeout_uuid):
        begin_time = time.time()
        res = await test_getevent_workflow("non-existent-uuid", "key1", 1.0)
        duration = time.time() - begin_time
        assert duration < 0.3
        assert res is None

    # No OAOO for getEvent outside of a workflow
    begin_time = time.time()
    res = await dbos.get_event_async("non-existent-uuid", "key1", 1.0)
    duration = time.time() - begin_time
    assert duration > 0.7
    assert res is None

    begin_time = time.time()
    res = await dbos.get_event_async("non-existent-uuid", "key1", 1.0)
    duration = time.time() - begin_time
    assert duration > 0.7
    assert res is None

    # Test setEvent outside of a workflow
    with pytest.raises(Exception) as exc_info:
        await dbos.set_event_async("key1", "value1")
    assert "set_event() must be called from within a workflow" in str(exc_info.value)


@pytest.mark.asyncio
async def test_sleep(dbos: DBOS) -> None:
    @DBOS.workflow()
    async def test_sleep_workflow(secs: float) -> str:
        await dbos.sleep_async(secs)
        return DBOS.workflow_id

    start_time = time.time()
    sleep_uuid = await test_sleep_workflow(1.5)
    assert time.time() - start_time > 1.4

    # Test sleep OAOO, skip sleep
    start_time = time.time()
    with SetWorkflowID(sleep_uuid):
        assert (await test_sleep_workflow(1.5)) == sleep_uuid
        assert time.time() - start_time < 0.3


def test_async_tx_raises(config: ConfigFile) -> None:

    with pytest.raises(DBOSException) as exc_info:

        @DBOS.transaction()
        async def test_async_tx() -> None:
            pass

    # destroy call needed to avoid "functions were registered but DBOS() was not called" warning
    DBOS.destroy(destroy_registry=True)


@pytest.mark.asyncio
async def test_async_step_temp(dbos: DBOS) -> None:
    step_counter: int = 0

    @DBOS.step()
    async def test_step(var: str) -> str:
        await asyncio.sleep(0.1)
        nonlocal step_counter
        step_counter += 1
        DBOS.logger.info("I'm test_step")
        return var + f"step{step_counter}"

    wfuuid = f"test_async_step_temp-{time.time_ns()}"
    with SetWorkflowID(wfuuid):
        result = await test_step("alice")
        assert result == "alicestep1"
    dbos._sys_db.wait_for_buffer_flush()

    with SetWorkflowID(wfuuid):
        result = await test_step("alice")
        assert result == "alicestep1"

    assert step_counter == 1



import threading
import uuid
from typing import Callable, Optional

import pytest
import sqlalchemy as sa

# Public API
from dbos import DBOS, DBOSConfiguredInstance, Queue, SetWorkflowID

# Private API used because this is a test
from dbos._context import DBOSContextEnsure, assert_current_dbos_context
from tests.conftest import queue_entries_are_cleaned_up


def test_required_roles(dbos: DBOS) -> None:
    @DBOS.required_roles(["user"])
    def tfunc(var: str) -> str:
        assert assert_current_dbos_context().assumed_role == "user"
        return var

    with pytest.raises(Exception) as exc_info:
        tfunc("bare-ctx")
    assert (
        str(exc_info.value)
        == "DBOS Error 8: Function tfunc requires a role, but was called in a context without authentication information"
    )

    with DBOSContextEnsure():
        with pytest.raises(Exception) as exc_info:
            tfunc("bare-ctx")
        assert (
            str(exc_info.value)
            == "DBOS Error 8: Function tfunc requires a role, but was called in a context without authentication information"
        )

        ctx = assert_current_dbos_context()
        ctx.authenticated_roles = ["a", "b", "c"]

        with pytest.raises(Exception) as exc_info:
            tfunc("bare-ctx")
        assert (
            str(exc_info.value)
            == "DBOS Error 8: Function tfunc has required roles, but user is not authenticated for any of them"
        )

        ctx.authenticated_roles = ["a", "b", "c", "user"]
        tfunc("bare-ctx")


def test_required_roles_class(dbos: DBOS) -> None:
    @DBOS.default_required_roles(["user"])
    class DBOSTestClassRR(DBOSConfiguredInstance):
        def __init__(self) -> None:
            super().__init__("myconfig")

        @DBOS.workflow()
        def test_func_user(self, var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "user"
            return var

        @DBOS.workflow()
        @DBOS.required_roles(["admin"])
        def test_func_admin(self, var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "admin"
            return var

        @DBOS.required_roles(["admin"])
        @DBOS.workflow()
        def test_func_admin_r(self, var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "admin"
            return var

        @classmethod
        @DBOS.workflow()
        def test_func_user_c(cls, var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "user"
            return var

        @classmethod
        @DBOS.workflow()
        @DBOS.required_roles(["admin"])
        def test_func_admin_c(cls, var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "admin"
            return var

        @classmethod
        @DBOS.required_roles(["admin"])
        @DBOS.workflow()
        def test_func_admin_r_c(cls, var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "admin"
            return var

        @staticmethod
        @DBOS.workflow()
        def test_func_user_s(var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "user"
            return var

        @staticmethod
        @DBOS.workflow()
        @DBOS.required_roles(["admin"])
        def test_func_admin_s(var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "admin"
            return var

        @staticmethod
        @DBOS.required_roles(["admin"])
        @DBOS.workflow()
        def test_func_admin_r_s(var: str) -> str:
            assert assert_current_dbos_context().assumed_role == "admin"
            return var

    # This checks that the interception occurs in all methods
    inst = DBOSTestClassRR()
    with pytest.raises(Exception) as exc_info:
        inst.test_func_user("inst-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)
    with pytest.raises(Exception) as exc_info:
        inst.test_func_admin("inst-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)
    with pytest.raises(Exception) as exc_info:
        inst.test_func_admin_r("inst-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)

    with pytest.raises(Exception) as exc_info:
        DBOSTestClassRR.test_func_user_c("class-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)
    with pytest.raises(Exception) as exc_info:
        DBOSTestClassRR.test_func_admin_c("inst-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)
    with pytest.raises(Exception) as exc_info:
        DBOSTestClassRR.test_func_admin_r_c("inst-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)

    with pytest.raises(Exception) as exc_info:
        DBOSTestClassRR.test_func_user_s("class-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)
    with pytest.raises(Exception) as exc_info:
        DBOSTestClassRR.test_func_admin_s("inst-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)
    with pytest.raises(Exception) as exc_info:
        DBOSTestClassRR.test_func_admin_r_s("inst-no-ctx")
    assert "DBOS Error 8" in str(exc_info.value)

    with DBOSContextEnsure():
        ctx = assert_current_dbos_context()
        # Admin is allowed all
        ctx.authenticated_roles = ["user", "admin"]

        inst.test_func_user("inst-no-ctx")
        inst.test_func_admin("inst-no-ctx")
        inst.test_func_admin_r("inst-no-ctx")

        DBOSTestClassRR.test_func_user_c("class-no-ctx")
        DBOSTestClassRR.test_func_admin_c("inst-no-ctx")
        DBOSTestClassRR.test_func_admin_r_c("inst-no-ctx")

        DBOSTestClassRR.test_func_user_s("class-no-ctx")
        DBOSTestClassRR.test_func_admin_s("inst-no-ctx")
        DBOSTestClassRR.test_func_admin_r_s("inst-no-ctx")

        # User is allowed some
        ctx.authenticated_roles = ["user"]

        inst.test_func_user("inst-no-ctx")
        DBOSTestClassRR.test_func_user_c("class-no-ctx")
        DBOSTestClassRR.test_func_user_s("class-no-ctx")

        # But not all...
        with pytest.raises(Exception) as exc_info:
            inst.test_func_admin("inst-no-ctx")
        assert "DBOS Error 8" in str(exc_info.value)
        with pytest.raises(Exception) as exc_info:
            inst.test_func_admin_r("inst-no-ctx")
        assert "DBOS Error 8" in str(exc_info.value)
        with pytest.raises(Exception) as exc_info:
            DBOSTestClassRR.test_func_admin_c("inst-no-ctx")
        assert "DBOS Error 8" in str(exc_info.value)
        with pytest.raises(Exception) as exc_info:
            DBOSTestClassRR.test_func_admin_r_c("inst-no-ctx")
        assert "DBOS Error 8" in str(exc_info.value)
        with pytest.raises(Exception) as exc_info:
            DBOSTestClassRR.test_func_admin_s("inst-no-ctx")
        assert "DBOS Error 8" in str(exc_info.value)
        with pytest.raises(Exception) as exc_info:
            DBOSTestClassRR.test_func_admin_r_s("inst-no-ctx")
        assert "DBOS Error 8" in str(exc_info.value)


# We can put classes in functions to test decorators for now...
def test_simple_workflow_static(dbos: DBOS) -> None:
    @DBOS.dbos_class()
    class DBOSTestClassStatic:
        txn_counter: int = 0
        wf_counter: int = 0
        step_counter: int = 0

        @staticmethod
        @DBOS.workflow()
        def test_workflow(var: str, var2: str) -> str:
            DBOSTestClassStatic.wf_counter += 1
            res = DBOSTestClassStatic.test_transaction(var2)
            res2 = DBOSTestClassStatic.test_step(var)
            DBOS.logger.info("I'm test_workflow")
            return res + res2

        @staticmethod
        @DBOS.transaction()
        def test_transaction(var2: str) -> str:
            rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
            DBOSTestClassStatic.txn_counter += 1
            DBOS.logger.info("I'm test_transaction")
            return var2 + str(rows[0][0])

        @staticmethod
        @DBOS.step()
        def test_step(var: str) -> str:
            DBOSTestClassStatic.step_counter += 1
            DBOS.logger.info("I'm test_step")
            return var

    assert DBOSTestClassStatic.test_workflow("bob", "bob") == "bob1bob"
    wfh = dbos.start_workflow(DBOSTestClassStatic.test_workflow, "bob", "bob")
    assert wfh.get_result() == "bob1bob"
    assert DBOSTestClassStatic.txn_counter == 2
    assert DBOSTestClassStatic.wf_counter == 2
    assert DBOSTestClassStatic.step_counter == 2


def test_simple_workflow_class(dbos: DBOS) -> None:
    @DBOS.dbos_class()
    class DBOSTestClassClass:
        txn_counter: int = 0
        wf_counter: int = 0
        step_counter: int = 0

        @classmethod
        @DBOS.workflow()
        def test_workflow(cls, var: str, var2: str) -> str:
            DBOSTestClassClass.wf_counter += 1
            res = DBOSTestClassClass.test_transaction(var2)
            res2 = DBOSTestClassClass.test_step(var)
            DBOS.logger.info("I'm test_workflow")
            return res + res2

        @classmethod
        @DBOS.transaction()
        def test_transaction(cls, var2: str) -> str:
            rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
            DBOSTestClassClass.txn_counter += 1
            DBOS.logger.info("I'm test_transaction")
            return var2 + str(rows[0][0])

        @classmethod
        @DBOS.step()
        def test_step(cls, var: str) -> str:
            DBOSTestClassClass.step_counter += 1
            DBOS.logger.info("I'm test_step")
            return var

    assert DBOSTestClassClass.test_workflow("bob", "bob") == "bob1bob"
    wfh = dbos.start_workflow(DBOSTestClassClass.test_workflow, "bob", "bob")
    assert wfh.get_result() == "bob1bob"
    assert DBOSTestClassClass.txn_counter == 2
    assert DBOSTestClassClass.wf_counter == 2
    assert DBOSTestClassClass.step_counter == 2


def test_no_instname(dbos: DBOS) -> None:
    @DBOS.dbos_class()
    class DBOSTestClassInst:
        @DBOS.workflow()
        def test_workflow(self) -> str:
            return "Nope"

    with pytest.raises(Exception) as exc_info:
        DBOSTestClassInst().test_workflow()
    assert "Function target appears to be a class instance, but does not have `config_name` set"


def test_simple_workflow_inst(dbos: DBOS) -> None:
    @DBOS.dbos_class()
    class DBOSTestClassInst(DBOSConfiguredInstance):
        def __init__(self) -> None:
            super().__init__("bob")
            self.txn_counter: int = 0
            self.wf_counter: int = 0
            self.step_counter: int = 0

        @DBOS.workflow()
        def test_workflow(self, var: str, var2: str) -> str:
            self.wf_counter += 1
            res = self.test_transaction(var2)
            res2 = self.test_step(var)
            DBOS.logger.info("I'm test_workflow")
            return res + res2

        @DBOS.transaction()
        def test_transaction(self, var2: str) -> str:
            rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
            self.txn_counter += 1
            DBOS.logger.info("I'm test_transaction")
            return var2 + str(rows[0][0])

        @DBOS.step()
        def test_step(self, var: str) -> str:
            self.step_counter += 1
            DBOS.logger.info("I'm test_step")
            return var

    inst = DBOSTestClassInst()
    assert inst.test_workflow("bob", "bob") == "bob1bob"

    wfh = dbos.start_workflow(inst.test_workflow, "bob", "bob")
    stati = dbos.get_workflow_status(wfh.get_workflow_id())
    assert stati
    assert stati.config_name == "bob"
    assert stati.class_name == "DBOSTestClassInst"
    stat = wfh.get_status()
    assert stat
    assert stat.config_name == "bob"
    assert stat.class_name == "DBOSTestClassInst"

    assert wfh.get_result() == "bob1bob"
    assert inst.txn_counter == 2
    assert inst.wf_counter == 2
    assert inst.step_counter == 2


def test_forgotten_decorator(dbos: DBOS) -> None:
    class DBOSTestRegErr(DBOSConfiguredInstance):
        def __init__(self) -> None:
            super().__init__("bob")
            self.txn_counter: int = 0
            self.wf_counter: int = 0
            self.step_counter: int = 0

        @DBOS.workflow()
        def test_workflow1(self) -> str:
            return "Forgot class decorator!"

        @classmethod
        @DBOS.workflow()
        def test_workflow2(cls) -> str:
            return "Forgot class decorator!"

    inst = DBOSTestRegErr()
    with pytest.raises(Exception) as exc_info:
        inst.test_workflow1()
    assert (
        "Function target appears to be a class instance, but is not properly registered"
        in str(exc_info.value)
    )

    with pytest.raises(Exception) as exc_info:
        DBOSTestRegErr.test_workflow2()

    assert (
        "Function target appears to be a class, but is not properly registered"
        in str(exc_info.value)
    )


def test_duplicate_reg(dbos: DBOS) -> None:
    @DBOS.dbos_class()
    class DBOSTestRegDup(DBOSConfiguredInstance):
        def __init__(self) -> None:
            super().__init__("bob")

    # Duplicate class registration
    with pytest.raises(Exception) as exc_info:

        @DBOS.dbos_class()
        class DBOSTestRegDup(DBOSConfiguredInstance):  # type: ignore
            def __init__(self) -> None:
                super().__init__("bob")

    assert "Duplicate type registration for class 'DBOSTestRegDup'" == str(
        exc_info.value
    )

    # Dupliocate instance registration
    inst = DBOSTestRegDup()
    with pytest.raises(Exception) as exc_info:
        inst = DBOSTestRegDup()

    assert (
        "Duplicate instance registration for class 'DBOSTestRegDup' instance 'bob'"
        == str(exc_info.value)
    )


def test_class_recovery(dbos: DBOS) -> None:
    exc_cnt: int = 0

    @DBOS.dbos_class()
    class DBOSTestClassRec:
        @classmethod
        @DBOS.workflow()
        def check_cls(cls, arg1: str) -> str:
            nonlocal exc_cnt
            exc_cnt += 1
            assert arg1 == "arg1"
            assert cls == DBOSTestClassRec
            return "ran"

    with SetWorkflowID("run1"):
        assert "ran" == DBOSTestClassRec.check_cls("arg1")

    assert exc_cnt == 1

    # Test we can execute the workflow by uuid as recovery would do
    handle = DBOS.execute_workflow_id("run1")
    assert handle.get_result() == "ran"
    assert exc_cnt == 2


def test_inst_recovery(dbos: DBOS) -> None:
    wfid = str(uuid.uuid4())
    exc_cnt: int = 0
    last_inst: Optional[TestClass] = None

    @DBOS.dbos_class()
    class TestClass(DBOSConfiguredInstance):
        def __init__(self) -> None:
            super().__init__("test_class")

        @DBOS.workflow()
        def check_inst(self, arg1: str) -> str:
            nonlocal exc_cnt
            nonlocal last_inst
            exc_cnt += 1
            assert arg1 == "arg1"
            last_inst = self
            return "ran2"

    inst = TestClass()
    with SetWorkflowID(wfid):
        assert "ran2" == inst.check_inst("arg1")

    assert exc_cnt == 1
    assert last_inst is inst

    # Test we can execute the workflow by uuid as recovery would do
    last_inst = None
    handle = DBOS.execute_workflow_id(wfid)
    assert handle.get_result() == "ran2"
    assert exc_cnt == 2
    assert last_inst is inst

    status = DBOS.retrieve_workflow(wfid).get_status()
    assert status.class_name == "TestClass"
    assert status.config_name == "test_class"


def test_inst_async_recovery(dbos: DBOS) -> None:
    wfid = str(uuid.uuid4())
    event = threading.Event()

    @DBOS.dbos_class()
    class TestClass(DBOSConfiguredInstance):

        def __init__(self, multiplier: int) -> None:
            self.multiply: Callable[[int], int] = lambda x: x * multiplier
            super().__init__("test_class")

        @DBOS.workflow()
        def workflow(self, x: int) -> int:
            event.wait()
            return self.multiply(x)

    input = 2
    multiplier = 5
    inst = TestClass(multiplier)

    with SetWorkflowID(wfid):
        orig_handle = DBOS.start_workflow(inst.workflow, input)

    status = orig_handle.get_status()
    assert status.class_name == "TestClass"
    assert status.config_name == "test_class"

    recovery_handle = DBOS.execute_workflow_id(wfid)

    event.set()
    assert orig_handle.get_result() == input * multiplier
    assert recovery_handle.get_result() == input * multiplier


def test_inst_async_step_recovery(dbos: DBOS) -> None:
    wfid = str(uuid.uuid4())
    event = threading.Event()

    @DBOS.dbos_class()
    class TestClass(DBOSConfiguredInstance):

        def __init__(self, multiplier: int) -> None:
            self.multiply: Callable[[int], int] = lambda x: x * multiplier
            super().__init__("test_class")

        @DBOS.step()
        def step(self, x: int) -> int:
            event.wait()
            return self.multiply(x)

    input = 2
    multiplier = 5
    inst = TestClass(multiplier)

    with SetWorkflowID(wfid):
        orig_handle = DBOS.start_workflow(inst.step, input)

    status = orig_handle.get_status()
    assert status.class_name == "TestClass"
    assert status.config_name == "test_class"

    recovery_handle = DBOS.execute_workflow_id(wfid)

    event.set()
    assert orig_handle.get_result() == input * multiplier
    assert recovery_handle.get_result() == input * multiplier


def test_step_recovery(dbos: DBOS) -> None:
    wfid = str(uuid.uuid4())
    thread_event = threading.Event()
    blocking_event = threading.Event()
    return_value = None

    @DBOS.dbos_class()
    class TestClass(DBOSConfiguredInstance):

        def __init__(self, multiplier: int) -> None:
            self.multiply: Callable[[int], int] = lambda x: x * multiplier
            super().__init__("test_class")

        @DBOS.step()
        def step(self, x: int) -> int:
            thread_event.set()
            blocking_event.wait()
            return self.multiply(x)

    input = 2
    multiplier = 5
    inst = TestClass(multiplier)

    # We're testing synchronously calling the step, but need to do so
    # asynchronously. Hence, a thread.
    def call_step() -> None:
        with SetWorkflowID(wfid):
            nonlocal return_value
            return_value = inst.step(input)

    thread = threading.Thread(target=call_step)
    thread.start()
    thread_event.wait()

    status = DBOS.retrieve_workflow(wfid).get_status()
    assert status.class_name == "TestClass"
    assert status.config_name == "test_class"

    recovery_handle = DBOS.execute_workflow_id(wfid)

    blocking_event.set()
    thread.join()
    assert return_value == input * multiplier
    assert recovery_handle.get_result() == input * multiplier


def test_class_queue_recovery(dbos: DBOS) -> None:
    step_counter: int = 0
    queued_steps = 5
    multiplier = 5

    wfid = str(uuid.uuid4())
    queue = Queue("test_queue")
    step_events = [threading.Event() for _ in range(queued_steps)]
    event = threading.Event()

    @DBOS.dbos_class()
    class TestClass(DBOSConfiguredInstance):
        def __init__(self, multiplier: int) -> None:
            self.multiply: Callable[[int], int] = lambda x: x * multiplier
            super().__init__("test_class")

        @DBOS.workflow()
        def test_workflow(self) -> list[int]:
            assert DBOS.workflow_id == wfid
            handles = []
            for i in range(queued_steps):
                h = queue.enqueue(self.test_step, i)
                handles.append(h)
            return [h.get_result() for h in handles]

        @DBOS.step()
        def test_step(self, i: int) -> int:
            nonlocal step_counter
            step_counter += 1
            step_events[i].set()
            event.wait()
            return self.multiply(i)

    inst = TestClass(multiplier)

    # Start the workflow. Wait for all five steps to start. Verify that they started.
    with SetWorkflowID(wfid):
        original_handle = DBOS.start_workflow(inst.test_workflow)
    for e in step_events:
        e.wait()
        e.clear()

    assert step_counter == 5

    # Recover the workflow, then resume it.
    recovery_handles = DBOS.recover_pending_workflows()
    # Wait until the 2nd invocation of the workflows are dequeued and executed
    for e in step_events:
        e.wait()
    event.set()

    # There should be one handle for the workflow and another for each queued step.
    assert len(recovery_handles) == queued_steps + 1
    # Verify that both the recovered and original workflows complete correctly.
    result = [i * multiplier for i in range(5)]
    for h in recovery_handles:
        status = h.get_status()
        assert status.class_name == "TestClass"
        assert status.config_name == "test_class"
        if h.get_workflow_id() == wfid:
            assert h.get_result() == result
    assert original_handle.get_result() == result
    # Each step should start twice, once originally and once in recovery.
    assert step_counter == 10

    # Rerun the workflow. Because each step is complete, none should start again.
    with SetWorkflowID(wfid):
        assert inst.test_workflow() == result
    assert step_counter == 10

    # Verify all queue entries eventually get cleaned up.
    assert queue_entries_are_cleaned_up(dbos)


def test_class_static_queue_recovery(dbos: DBOS) -> None:
    step_counter: int = 0
    queued_steps = 5

    wfid = str(uuid.uuid4())
    queue = Queue("test_queue")
    step_events = [threading.Event() for _ in range(queued_steps)]
    event = threading.Event()

    @DBOS.dbos_class()
    class TestClass:
        @staticmethod
        @DBOS.workflow()
        def test_workflow() -> list[int]:
            assert DBOS.workflow_id == wfid
            handles = []
            for i in range(queued_steps):
                h = queue.enqueue(TestClass.test_step, i)
                handles.append(h)
            return [h.get_result() for h in handles]

        @staticmethod
        @DBOS.step()
        def test_step(i: int) -> int:
            nonlocal step_counter
            step_counter += 1
            step_events[i].set()
            event.wait()
            return i

    # Start the workflow. Wait for all five steps to start. Verify that they started.
    with SetWorkflowID(wfid):
        original_handle = DBOS.start_workflow(TestClass.test_workflow)
    for e in step_events:
        e.wait()
        e.clear()

    assert step_counter == 5

    # Recover the workflow, then resume it.
    recovery_handles = DBOS.recover_pending_workflows()
    # Wait until the 2nd invocation of the workflows are dequeued and executed
    for e in step_events:
        e.wait()
    event.set()

    # There should be one handle for the workflow and another for each queued step.
    assert len(recovery_handles) == queued_steps + 1
    # Verify that both the recovered and original workflows complete correctly.
    result = [i for i in range(5)]
    for h in recovery_handles:
        status = h.get_status()
        # Class name is not recorded for static methods
        assert status.class_name == None
        assert status.config_name == None
        if h.get_workflow_id() == wfid:
            assert h.get_result() == result
    assert original_handle.get_result() == result
    # Each step should start twice, once originally and once in recovery.
    assert step_counter == 10

    # Rerun the workflow. Because each step is complete, none should start again.
    with SetWorkflowID(wfid):
        assert TestClass.test_workflow() == result
    assert step_counter == 10

    # Verify all queue entries eventually get cleaned up.
    assert queue_entries_are_cleaned_up(dbos)


def test_class_classmethod_queue_recovery(dbos: DBOS) -> None:
    step_counter: int = 0
    multiplier = 5
    queued_steps = 5

    wfid = str(uuid.uuid4())
    queue = Queue("test_queue")
    step_events = [threading.Event() for _ in range(queued_steps)]
    event = threading.Event()

    @DBOS.dbos_class()
    class TestClass:
        multiply: Callable[[int], int] = lambda _: 0

        @classmethod
        @DBOS.workflow()
        def test_workflow(cls) -> list[int]:
            cls.multiply = lambda x: x * multiplier
            assert DBOS.workflow_id == wfid
            handles = []
            for i in range(queued_steps):
                h = queue.enqueue(TestClass.test_step, i)
                handles.append(h)
            return [h.get_result() for h in handles]

        @classmethod
        @DBOS.step()
        def test_step(cls, i: int) -> int:
            nonlocal step_counter
            step_counter += 1
            step_events[i].set()
            event.wait()
            return cls.multiply(i)

    # Start the workflow. Wait for all five steps to start. Verify that they started.
    with SetWorkflowID(wfid):
        original_handle = DBOS.start_workflow(TestClass.test_workflow)
    for e in step_events:
        e.wait()
        e.clear()

    assert step_counter == 5

    # Recover the workflow, then resume it.
    recovery_handles = DBOS.recover_pending_workflows()
    # Wait until the 2nd invocation of the workflows are dequeued and executed
    for e in step_events:
        e.wait()
    event.set()

    # There should be one handle for the workflow and another for each queued step.
    assert len(recovery_handles) == queued_steps + 1
    # Verify that both the recovered and original workflows complete correctly.
    result = [i * multiplier for i in range(5)]
    for h in recovery_handles:
        status = h.get_status()
        # Class name is recorded for class methods
        assert status.class_name == "TestClass"
        assert status.config_name == None
        if h.get_workflow_id() == wfid:
            assert h.get_result() == result
    assert original_handle.get_result() == result
    # Each step should start twice, once originally and once in recovery.
    assert step_counter == 10

    # Rerun the workflow. Because each step is complete, none should start again.
    with SetWorkflowID(wfid):
        assert TestClass.test_workflow() == result
    assert step_counter == 10

    # Verify all queue entries eventually get cleaned up.
    assert queue_entries_are_cleaned_up(dbos)


def test_inst_txn(dbos: DBOS) -> None:
    wfid = str(uuid.uuid4())

    @DBOS.dbos_class()
    class TestClass(DBOSConfiguredInstance):

        def __init__(self, multiplier: int) -> None:
            self.multiply: Callable[[int], int] = lambda x: x * multiplier
            super().__init__("test_class")

        @DBOS.transaction()
        def transaction(self, x: int) -> int:
            return self.multiply(x)

    input = 2
    multiplier = 5
    inst = TestClass(multiplier)

    with SetWorkflowID(wfid):
        assert inst.transaction(input) == input * multiplier
    dbos._sys_db.wait_for_buffer_flush()
    status = DBOS.retrieve_workflow(wfid).get_status()
    assert status.class_name == "TestClass"
    assert status.config_name == "test_class"

    handle = DBOS.start_workflow(inst.transaction, input)
    assert handle.get_result() == input * multiplier
    dbos._sys_db.wait_for_buffer_flush()
    status = handle.get_status()
    assert status.class_name == "TestClass"
    assert status.config_name == "test_class"


def test_mixed_methods(dbos: DBOS) -> None:

    @DBOS.dbos_class()
    class TestClass(DBOSConfiguredInstance):

        def __init__(self, multiplier: int) -> None:
            self.multiply: Callable[[int], int] = lambda x: x * multiplier
            super().__init__("test_class")

        @DBOS.workflow()
        def instance_workflow(self, x: int) -> int:
            return self.multiply(x)

        @classmethod
        @DBOS.workflow()
        def classmethod_workflow(cls, x: int) -> int:
            return x

        @staticmethod
        @DBOS.workflow()
        def staticmethod_workflow(x: int) -> int:
            return x

    input = 2
    multiplier = 5
    inst = TestClass(multiplier)

    handle = DBOS.start_workflow(inst.instance_workflow, input)
    assert handle.get_result() == input * multiplier
    status = handle.get_status()
    assert status.class_name == "TestClass"
    assert status.config_name == "test_class"

    handle = DBOS.start_workflow(inst.classmethod_workflow, input)
    assert handle.get_result() == input
    status = handle.get_status()
    assert status.class_name == "TestClass"
    assert status.config_name == None

    handle = DBOS.start_workflow(inst.staticmethod_workflow, input)
    assert handle.get_result() == input
    status = handle.get_status()
    assert status.class_name == None
    assert status.config_name == None



import datetime
import threading
import time
import uuid

import pytest
import sqlalchemy as sa
from psycopg.errors import SerializationFailure
from sqlalchemy.exc import InvalidRequestError, OperationalError

# Public API
from dbos import DBOS, GetWorkflowsInput, SetWorkflowID
from dbos._error import DBOSDeadLetterQueueError, DBOSException
from dbos._sys_db import WorkflowStatusString


def test_transaction_errors(dbos: DBOS) -> None:
    retry_counter: int = 0

    @DBOS.transaction()
    def test_retry_transaction(max_retry: int) -> int:
        nonlocal retry_counter
        if retry_counter < max_retry:
            retry_counter += 1
            raise OperationalError(
                "Serialization test error", {}, SerializationFailure()
            )
        return max_retry

    @DBOS.transaction()
    def test_noretry_transaction() -> None:
        nonlocal retry_counter
        retry_counter += 1
        DBOS.sql_session.execute(sa.text("selct abc from c;")).fetchall()

    res = test_retry_transaction(10)
    assert res == 10
    assert retry_counter == 10

    with pytest.raises(Exception) as exc_info:
        test_noretry_transaction()
    assert exc_info.value.orig.sqlstate == "42601"  # type: ignore
    assert retry_counter == 11


def test_invalid_transaction_error(dbos: DBOS) -> None:
    commit_txn_counter: int = 0
    rollback_txn_counter: int = 0

    @DBOS.transaction()
    def test_commit_transaction() -> None:
        nonlocal commit_txn_counter
        commit_txn_counter += 1
        # Commit shouldn't be allowed to be called in a transaction. The error message should be clear.
        DBOS.sql_session.commit()
        return

    @DBOS.transaction()
    def test_abort_transaction() -> None:
        nonlocal rollback_txn_counter
        rollback_txn_counter += 1
        # Rollback shouldn't be allowed to be called in a transaction. The error message should be clear.
        DBOS.sql_session.rollback()
        return

    # Test OAOO and exception handling
    wfuuid = str(uuid.uuid4())
    with pytest.raises(InvalidRequestError) as exc_info:
        with SetWorkflowID(wfuuid):
            test_commit_transaction()
    assert "Can't operate on closed transaction inside context manager." in str(
        exc_info.value
    )
    print(exc_info.value)

    with pytest.raises(InvalidRequestError) as exc_info:
        with SetWorkflowID(wfuuid):
            test_commit_transaction()
    assert "Can't operate on closed transaction inside context manager." in str(
        exc_info.value
    )

    assert commit_txn_counter == 1

    wfuuid = str(uuid.uuid4())
    with pytest.raises(InvalidRequestError) as exc_info:
        with SetWorkflowID(wfuuid):
            test_abort_transaction()
    assert "Can't operate on closed transaction inside context manager." in str(
        exc_info.value
    )
    print(exc_info.value)

    with pytest.raises(InvalidRequestError) as exc_info:
        with SetWorkflowID(wfuuid):
            test_abort_transaction()
    assert "Can't operate on closed transaction inside context manager." in str(
        exc_info.value
    )
    assert rollback_txn_counter == 1


def test_notification_errors(dbos: DBOS) -> None:
    @DBOS.workflow()
    def test_send_workflow(dest_uuid: str, topic: str) -> str:
        DBOS.send(dest_uuid, "test1")
        DBOS.send(dest_uuid, "test2", topic=topic)
        DBOS.send(dest_uuid, "test3")
        return dest_uuid

    @DBOS.workflow()
    def test_recv_workflow(topic: str) -> str:
        msg1 = DBOS.recv(topic, timeout_seconds=10)
        msg2 = DBOS.recv(timeout_seconds=10)
        msg3 = DBOS.recv(timeout_seconds=10)
        return "-".join([str(msg1), str(msg2), str(msg3)])

    # Crash the notification connection and make sure send/recv works on time.
    while dbos._sys_db.notification_conn is None:
        time.sleep(1)
    dbos._sys_db.notification_conn.close()
    assert dbos._sys_db.notification_conn.closed == 1

    # Wait for the connection to be re-established
    while dbos._sys_db.notification_conn.closed != 0:
        time.sleep(1)

    dest_uuid = str("sruuid1")
    with SetWorkflowID(dest_uuid):
        handle = dbos.start_workflow(test_recv_workflow, "testtopic")
        assert handle.get_workflow_id() == dest_uuid

    send_uuid = str("sruuid2")
    with SetWorkflowID(send_uuid):
        res = test_send_workflow(handle.get_workflow_id(), "testtopic")
        assert res == dest_uuid

    begin_time = time.time()
    assert handle.get_result() == "test2-test1-test3"
    duration = time.time() - begin_time
    assert duration < 3.0


def test_buffer_flush_errors(dbos: DBOS) -> None:
    @DBOS.transaction()
    def test_transaction(var: str) -> str:
        rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
        return var + str(rows[0][0])

    cur_time: str = datetime.datetime.now().isoformat()
    gwi: GetWorkflowsInput = GetWorkflowsInput()
    gwi.start_time = cur_time

    res = test_transaction("bob")
    assert res == "bob1"

    dbos._sys_db.wait_for_buffer_flush()
    wfs = dbos._sys_db.get_workflows(gwi)
    assert len(wfs.workflow_uuids) == 1

    # Crash the system database connection and make sure the buffer flush works on time.
    backup_engine = dbos._sys_db.engine
    dbos._sys_db.engine = sa.create_engine(
        "postgresql+psycopg://fake:database@localhost/fake_db"
    )

    res = test_transaction("bob")
    assert res == "bob1"

    # Should see some errors in the logs
    time.sleep(2)

    # Switch back to the original good engine.
    dbos._sys_db.engine = backup_engine

    dbos._sys_db.wait_for_buffer_flush()
    wfs = dbos._sys_db.get_workflows(gwi)
    assert len(wfs.workflow_uuids) == 2


def test_dead_letter_queue(dbos: DBOS) -> None:
    event = threading.Event()
    max_recovery_attempts = 20
    recovery_count = 0

    @DBOS.workflow(max_recovery_attempts=max_recovery_attempts)
    def dead_letter_workflow() -> None:
        nonlocal recovery_count
        recovery_count += 1
        event.wait()

    # Start a workflow that blocks forever
    wfid = str(uuid.uuid4())
    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(dead_letter_workflow)

    # Attempt to recover the blocked workflow the maximum number of times
    for i in range(max_recovery_attempts):
        DBOS.recover_pending_workflows()
        assert recovery_count == i + 2

    # Verify an additional attempt (either through recovery or through a direct call) throws a DLQ error
    # and puts the workflow in the DLQ status.
    with pytest.raises(Exception) as exc_info:
        DBOS.recover_pending_workflows()
    assert exc_info.errisinstance(DBOSDeadLetterQueueError)
    assert handle.get_status().status == WorkflowStatusString.RETRIES_EXCEEDED.value
    with pytest.raises(Exception) as exc_info:
        with SetWorkflowID(wfid):
            dead_letter_workflow()
    assert exc_info.errisinstance(DBOSDeadLetterQueueError)

    # Resume the workflow. Verify it returns to PENDING status without error.
    resumed_handle = dbos.resume_workflow(wfid)
    assert (
        handle.get_status().status
        == resumed_handle.get_status().status
        == WorkflowStatusString.PENDING.value
    )

    # Verify the workflow can recover again without error.
    DBOS.recover_pending_workflows()

    # Complete the blocked workflow
    event.set()
    assert handle.get_result() == resumed_handle.get_result() == None
    dbos._sys_db.wait_for_buffer_flush()
    assert handle.get_status().status == WorkflowStatusString.SUCCESS.value


def test_wfstatus_invalid(dbos: DBOS) -> None:
    @DBOS.workflow()
    def regular_workflow() -> str:
        return "done"

    has_executed = False

    @DBOS.workflow()
    def non_deterministic_worklow() -> None:
        nonlocal has_executed
        handle = dbos.start_workflow(regular_workflow)
        if not has_executed:
            # Mock a scenario where the workflow control flow is changed by an external process
            dbos.set_event("test_event", "value1")
            has_executed = True

        res = handle.get_result()
        assert res == "done"

        handle.get_status()
        return

    wfuuid = str(uuid.uuid4())
    with SetWorkflowID(wfuuid):
        non_deterministic_worklow()

    with pytest.raises(DBOSException) as exc_info:
        with SetWorkflowID(wfuuid):
            non_deterministic_worklow()
    assert "Hint: Check if your workflow is deterministic." in str(exc_info.value)



"""
Fix job queue PK.

Revision ID: 50f3227f0b4b
Revises: eab0cc1d9a14
Create Date: 2024-09-25 14:03:53.308068

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "50f3227f0b4b"
down_revision: Union[str, None] = "eab0cc1d9a14"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.drop_constraint("job_queue_pkey", "job_queue", schema="dbos", type_="primary")

    op.create_primary_key(
        "job_queue_pkey", "job_queue", ["workflow_uuid"], schema="dbos"
    )


def downgrade() -> None:
    # Reverting the changes
    op.drop_constraint("job_queue_pkey", "job_queue", schema="dbos", type_="primary")

    op.create_primary_key(
        "job_queue_pkey", "job_queue", ["created_at_epoch_ms"], schema="dbos"
    )



import uuid

import pytest
import sqlalchemy as sa
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

# Public API
from dbos import DBOS, SetWorkflowID


# Declare a SQLAlchemy ORM base class
class Base(DeclarativeBase):
    pass


# Declare a SQLAlchemy ORM class for accessing the database table.
class Hello(Base):
    __tablename__ = "dbos_hello"
    greet_count: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(nullable=False)

    def __repr__(self) -> str:
        return f"Hello(greet_count={self.greet_count!r}, name={self.name!r})"


def test_simple_transaction(dbos: DBOS, postgres_db_engine: sa.Engine) -> None:
    txn_counter: int = 0
    assert dbos._app_db_field is not None
    Base.metadata.drop_all(dbos._app_db_field.engine)
    Base.metadata.create_all(dbos._app_db_field.engine)

    @DBOS.transaction()
    def test_transaction(name: str) -> str:
        new_greeting = Hello(name=name)
        DBOS.sql_session.add(new_greeting)
        stmt = (
            sa.select(Hello)
            .where(Hello.name == name)
            .order_by(Hello.greet_count.desc())
            .limit(1)
        )
        row = DBOS.sql_session.scalar(stmt)
        assert row is not None
        greet_count = row.greet_count
        nonlocal txn_counter
        txn_counter += 1
        return name + str(greet_count)

    assert test_transaction("alice") == "alice1"
    assert test_transaction("alice") == "alice2"
    assert txn_counter == 2

    # Test OAOO
    wfuuid = str(uuid.uuid4())
    with SetWorkflowID(wfuuid):
        assert test_transaction("alice") == "alice3"
    with SetWorkflowID(wfuuid):
        assert test_transaction("alice") == "alice3"
    assert txn_counter == 3  # Only increment once

    Base.metadata.drop_all(dbos._app_db_field.engine)

    # Make sure no transactions are left open
    with postgres_db_engine.begin() as conn:
        result = conn.execute(
            sa.text(
                "select * from pg_stat_activity where state = 'idle in transaction'"
            )
        ).fetchall()
        assert len(result) == 0


def test_error_transaction(dbos: DBOS, postgres_db_engine: sa.Engine) -> None:
    txn_counter: int = 0
    assert dbos._app_db_field is not None
    # Drop the database but don't re-create. Should fail.
    Base.metadata.drop_all(dbos._app_db_field.engine)

    @DBOS.transaction()
    def test_transaction(name: str) -> str:
        nonlocal txn_counter
        txn_counter += 1
        new_greeting = Hello(name=name)
        DBOS.sql_session.add(new_greeting)
        return name

    with pytest.raises(Exception) as exc_info:
        test_transaction("alice")
    assert 'relation "dbos_hello" does not exist' in str(exc_info.value)
    assert txn_counter == 1

    # Test OAOO
    wfuuid = str(uuid.uuid4())
    with SetWorkflowID(wfuuid):
        with pytest.raises(Exception) as exc_info:
            test_transaction("alice")
    assert 'relation "dbos_hello" does not exist' in str(exc_info.value)
    assert txn_counter == 2

    with SetWorkflowID(wfuuid):
        with pytest.raises(Exception) as exc_info:
            test_transaction("alice")
    assert 'relation "dbos_hello" does not exist' in str(exc_info.value)
    assert txn_counter == 2

    # Make sure no transactions are left open
    with postgres_db_engine.begin() as conn:
        result = conn.execute(
            sa.text(
                "select * from pg_stat_activity where state = 'idle in transaction'"
            )
        ).fetchall()
        assert len(result) == 0
